open Util
open Ids

type script = condition_response list
and condition_response = condition * response list 
and condition = trigger list 
and trigger = {
  mutable trigger_id : Int32.t ;
  mutable negated : bool ;
  mutable t_1 : Int32.t ;
  mutable t_2 : Int32.t ;
  mutable t_3 : string ;
  mutable t_4 : string ;
  mutable t_5 : object_param ;
  unknown : Int32.t;
} 
and response = int * action list 
and action = {
  mutable action_id : Int32.t;
  mutable a_1 : object_param ; (* ActionOverride(a_1, ...) *)
  mutable a_2 : object_param ;
  mutable a_3 : object_param ;
  mutable a_4 : Int32.t ;
  a_5 : point_param; 
  a_6 : Int32.t ;
  a_7 : Int32.t ;
  mutable a_8 : string ;
  a_9 : string ;
} 
and object_param = {
  mutable o_ea : Int32.t ;
  mutable o_general : Int32.t ;
  mutable o_race : Int32.t ;
  mutable o_subrace : Int32.t ;
  mutable o_class : Int32.t ;
  mutable o_specific : Int32.t ;
  mutable o_gender : Int32.t ;
  mutable o_alignment : Int32.t ;
  mutable o_identifiers : Int32.t ;
  mutable o_unknown1 : Int32.t ;
  mutable o_unknown2 : Int32.t ;
  mutable o_unknown3 : Int32.t ;
  mutable o_unknown4 : Int32.t ;
  mutable o_name : string ;
  mutable o_rect : rect_param ; 
  mutable o_iwd2_1 : Int32.t ;
  mutable o_iwd2_2 : Int32.t ;
} 
and point_param = Int32.t * Int32.t 
and rect_param = Int32.t * Int32.t * Int32.t * Int32.t
let empty_rect () = 
  let n = Int32.of_int (-1) in
  (n , n , n , n)

let empty_object_param () = {
  o_ea = Int32.zero;
  o_general = Int32.zero;
  o_race = Int32.zero;
  o_subrace = Int32.zero;
  o_class = Int32.zero;
  o_specific = Int32.zero;
  o_gender = Int32.zero;
  o_alignment = Int32.zero;
  o_identifiers = Int32.zero;
  o_unknown1 = Int32.zero;
  o_unknown2 = Int32.zero;
  o_unknown3 = Int32.zero;
  o_unknown4 = Int32.zero;
  o_name = "";
  o_rect = empty_rect () ; 
  o_iwd2_1 = Int32.zero ;
  o_iwd2_2 = Int32.zero ; 
} 

let empty_action () = { 
  action_id = Int32.zero ;
  a_1 = empty_object_param () ; 
  a_2 = empty_object_param () ; 
  a_3 = empty_object_param () ; 
  a_4 = Int32.zero  ; 
  a_5 = (Int32.zero,Int32.zero) ; 
  a_6 = Int32.zero  ; 
  a_7 = Int32.zero  ; 
  a_8 = "";
  a_9 = "";
  } 

let empty_trigger () = { 
   trigger_id = Int32.zero ;
   negated = false ;
   t_1 = Int32.zero ;
   t_2 = Int32.zero ;
   t_3 = "" ;
   t_4 = "" ;
   t_5 = empty_object_param () ;
  unknown = Int32.zero ; 
} 

let ids_file_of_object_position iwd2 i = 
  if iwd2 then 
    match i with
    | 0 -> "EA"
    | 1 -> "GENERAL"
    | 2 -> "RACE" 
    | 3 -> "SUBRACE" 
    | 4 -> "CLASS" 
    | 5 -> "SPECIFIC" 
    | 6 -> "GENDER"
    | 7 -> "ALIGNMNT" 
    | _ -> failwith (Printf.sprintf "Don't know what to do with the %dth specifier" i )
  else match i with
  | 0 -> "EA"
  | 1 -> "GENERAL"
  | 2 -> "RACE" 
  | 3 -> "CLASS" 
  | 4 -> "SPECIFIC"
  | 5 -> "GENDER" 
  | 6 -> "ALIGN" 
  | _ -> failwith (Printf.sprintf "Don't know what to do with the %dth specifier" i )

type ids_map = { (* represents one IDS file *)
  from_int : (Int32.t, ids) Hashtbl.t ;
  from_sym : (string, ids) Hashtbl.t ;
  from_uppercase_sym : (string, ids) Hashtbl.t ;
} 

let non_empty_object o =
  o.o_ea <> Int32.zero ||
  o.o_general <> Int32.zero ||
  o.o_race <> Int32.zero ||
  o.o_class <> Int32.zero ||
  o.o_specific <> Int32.zero ||
  o.o_gender <> Int32.zero ||
  o.o_alignment <> Int32.zero ||
  o.o_identifiers <> Int32.zero ||
  o.o_unknown1 <> Int32.zero ||
  o.o_unknown2 <> Int32.zero ||
  o.o_unknown3 <> Int32.zero ||
  o.o_unknown4 <> Int32.zero ||
  o.o_name <> ""

type how_to_save_bcs = 
  | Save_BCS_OC of out_channel
  | Save_BCS_Buffer of Buffer.t 

let save_bcs game how bcs =
  let bcs_printf fmt = 
    let k result = match how with 
    | Save_BCS_OC(oc) -> output_string oc result
    | Save_BCS_Buffer(b) -> Buffer.add_string b result
    in 
    kprintf k fmt 
  in 
  let rec save_bcs game bcs =
    bcs_printf "SC\n" ;
    List.iter (save_cr game) bcs ;
    bcs_printf "SC\n"
  and save_cr game (c,rl) =
    bcs_printf "CR\nCO\n" ;
    List.iter (save_t game ) c ;
    bcs_printf "CO\nRS\n" ;
    List.iter (save_r game ) rl ;
    bcs_printf "RS\nCR\n"
  and save_t game t =
    bcs_printf "TR\n%ld %ld %d %ld %ld \"%s\" \"%s\" "
      t.trigger_id t.t_1 (if t.negated then 1 else 0)
      t.t_2 t.unknown t.t_3 t.t_4 ;
    save_obj game t.t_5 ;
    bcs_printf "TR\n" 
  and save_r game (w,al) =
    bcs_printf "RE\n%d" w ;
    List.iter (save_a game) al ;
    bcs_printf "RE\n" ;
  and save_a game a =
    bcs_printf "AC\n%ld" a.action_id;
    save_obj game a.a_1 ;
    save_obj game a.a_2 ;
    save_obj game  a.a_3 ;
    let p1,p2 = a.a_5 in 
    bcs_printf "%ld %ld %ld %ld %ld\"%s\" \"%s\" AC\n"
      a.a_4 p1 p2 a.a_6 a.a_7 a.a_8 a.a_9 
  and save_obj game obj =
    if (game.Load.is_iwd2) then 
      let w,x,y,z = obj.o_rect in 
    bcs_printf "OB\n%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld [%ld.%ld.%ld.%ld] \"%s\" %ld %ld OB\n"
    (*
      obj.o_ea
      obj.o_general
      obj.o_race
      obj.o_subrace
      obj.o_class
      obj.o_kit
      obj.o_specifics
      obj.o_instance
      obj.o_specialcase
      obj.o_classmask
      *)
      obj.o_ea
      obj.o_general
      obj.o_race
      Int32.zero (* obj.o_subrace *) (* NO CLUE JONES! THIS DOES NOT WORK *) 
      obj.o_specific
      obj.o_gender
      obj.o_alignment
      Int32.zero (* obj.o_class *) (* NO CLUE JONES! THIS DOES NOT WORK *) 
      obj.o_identifiers obj.o_unknown1 obj.o_unknown2
      obj.o_unknown3 obj.o_unknown4 
      w x y z 
      obj.o_name
      obj.o_iwd2_1 obj.o_iwd2_2 
    else 
    bcs_printf "OB\n%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld \"%s\"OB\n"
      obj.o_ea obj.o_general obj.o_race obj.o_class
      obj.o_specific obj.o_gender obj.o_alignment 
      obj.o_identifiers obj.o_unknown1 obj.o_unknown2
      obj.o_unknown3 obj.o_unknown4 obj.o_name
  in
  save_bcs game bcs 

let all_games_ids_state = Hashtbl.create 3 

(* let ids_state : (string, ids_map) Hashtbl.t = Hashtbl.create 511  *)

let make_ids_map il = 
  let from_int = Hashtbl.create 511 in
  let from_sym = Hashtbl.create 511 in
  let from_uppercase_sym = Hashtbl.create 511 in
  List.iter (fun ids ->
    (* log_and_print "IDS: %ld %s\n" ids.i_num ids.i_name ;  *)
    Hashtbl.add from_int ids.i_num ids ;
    Hashtbl.add from_sym ids.i_name ids ;
    Hashtbl.add from_uppercase_sym (String.uppercase ids.i_name) ids ;
  ) il ;
  { from_int = from_int ;
    from_sym = from_sym ;
    from_uppercase_sym = from_uppercase_sym; } 

let get_ids_map game ids_filename =
  try
    let ids_state = Hashtbl.find all_games_ids_state game.Load.game_path in
    Hashtbl.find ids_state ids_filename
  with e -> begin
    let buff,path = Load.load_resource game true ids_filename "IDS" in 
    let where_newline = String.index buff '\n' in
    let buff = Str.string_after buff where_newline in 
    let lexbuf = lex_init_from_string ids_filename buff in 
    let lexer = if game.Load.is_iwd2 && ids_filename <> "ACTION" &&
                ids_filename <> "TRIGGER" then 
          Idslexer2.initial 
      else Idslexer.initial 
    in 
    let result = Stats.time "parsing .ids files" (fun () ->
      Idsparser.ids_file lexer lexbuf) () in
    pop_context () ; 
    log_or_print "[%s.IDS] parsed\n" ids_filename ;
    let ids_map = make_ids_map result in 
    let my_ids_state = try 
      Hashtbl.find all_games_ids_state game.Load.game_path 
      with Not_found -> Hashtbl.create 511 
    in 
    Hashtbl.add my_ids_state ids_filename ids_map ;
    Hashtbl.add all_games_ids_state game.Load.game_path my_ids_state ;
    ids_map
  end 

let ids_of_int game ids_file i = 
  let ids_map : ids_map = get_ids_map game ids_file in
  (Hashtbl.find ids_map.from_int i)

let ids_of_sym game ids_file sym = 
  let ids_map : ids_map = get_ids_map game ids_file in
  try 
    (Hashtbl.find ids_map.from_sym sym)
  with Not_found ->
    let a = (Hashtbl.find ids_map.from_uppercase_sym (String.uppercase sym)) in
(*    let msg = Printf.sprintf "[%s] should be [%s] (note case)\n" 
      sym a.i_name in
    (try input_error "PARSE" msg with _ -> () ) ; *)
    a 

let every_ids_of_int game ids_file i = 
  let ids_map : ids_map = get_ids_map game ids_file in
  (Hashtbl.find_all ids_map.from_int i)

let int_of_sym game ids_file sym =
  let ids = ids_of_sym game ids_file sym in
  ids.i_num 

let sym_of_int game ids_file i =
  try
    let ids_map : ids_map = get_ids_map game (String.uppercase ids_file ) in
    (Hashtbl.find ids_map.from_int i).i_name
  with Not_found ->
    (* log_or_print "WARNING: %d (0x%x) not found in %s.IDS\n" i i ids_file ; *)
    Printf.sprintf "%ld" i 

let is_concat_string ids =
  match Int32.to_int ids.i_num with
  |  0x400F (*Global(S:Name.,S:Area.,I:Value.)*)
  |  0x4034 (*GlobalGT(S:Name.,S:Area.,I:Value.)*)
  |  0x4035 (*GlobalLT(S:Name.,S:Area.,I:Value.)*)
  |  30 (*SetGlobal(S:Name.,S:Area.,I:Value.)*)
  |  115 (*SetGlobalTimer(S:Name.,S:Area.,I:Time.GTimes)*)
  |  109 (*IncrementGlobal(S:Name.,S:Area.,I:Value.)*)
  |  246 (*CreateCreatureAtLocation(S:GLOBAL.,S:Area.,S:ResRef.)*)
  |  256 (*CreateItemGlobal(S:Global.,S:Area.,S:ResRef.)*)
  |  268 (* RealSetGlobalTimer(S:Name*,S:Area*,I:Time*GTimes) *)
  | 335 (* SetTokenGlobal(S:GLOBAL.,S:Area.,S:Token.) *)
      -> true 
  | _ -> false

let split6 s =
  if String.length s >= 6 then 
    (String.sub s 6 ((String.length s) - 6) ,
     String.sub s 0 6 )
  else
    "",s

let rec best_ids_of_trigger game c =
  let ids = every_ids_of_int game "TRIGGER" c.trigger_id in 
  if ids = [] then begin
    log_and_print "Cannot resolve trigger 0x%lx\n" c.trigger_id ;
    raise Not_found
  end else begin
  let rec proc lst = match lst with
  | [] -> (try List.hd ids with e -> 
    if c.trigger_id < Int32.of_int 0x4000 then begin
      c.trigger_id <- Int32.add c.trigger_id (Int32.of_int 0x4000) ;
      best_ids_of_trigger game c ;
    end else begin
    log_and_print "Cannot resolve trigger 0x%lx\n" c.trigger_id ; raise e
    end)
  | ids :: tl -> 
    let string_formal_count = List.fold_left  
      (fun acc elt  -> acc + 
        if (elt.arg_kind = Arg_String) then 1 else 0) 0 ids.i_args in
    if (string_formal_count > 0 && c.t_3 = "") ||
       (string_formal_count > 1 && c.t_4 = "") || 
       (string_formal_count = 0 && c.t_3 <> "") ||
       (string_formal_count = 0 && c.t_4 <> "") then
       proc tl
    else begin
      ids 
    end 
  in 
  proc ids
  end

let rec best_ids_of_action game a = 
  let ids = every_ids_of_int game "ACTION" a.action_id in 
  if ids = [] then begin
    log_and_print "Cannot resolve action 0x%lx\n" a.action_id ;
    raise Not_found
  end else begin
  let rec proc lst = match lst with
  | [] -> (try List.hd ids with e -> 
    if a.action_id < Int32.of_int 0x4000 then begin
      a.action_id <- Int32.add a.action_id (Int32.of_int 0x4000);
      best_ids_of_action game a ;
    end else begin 
    log_and_print "Cannot resolve action %ld\n" a.action_id ; raise e
    end)
  | ids :: tl -> 
    let string_formal_count = List.fold_left  
      (fun acc elt  -> acc + 
        if (elt.arg_kind = Arg_String) then 1 else 0) 0 ids.i_args in
    if (string_formal_count > 0 && a.a_8 = "") ||
       (string_formal_count > 1 && a.a_9 = "") || 
       (string_formal_count = 0 && a.a_8 <> "") ||
       (string_formal_count = 0 && a.a_9 <> "") then
       proc tl
    else begin
      ids 
    end 
  in 
  proc ids
  end

type aarg = 
  Act_Integer of Int32.t
| Act_String of string
| Act_Object of object_param 
| Act_Point of point_param 

type print_what =
| BCS_Print_Script of script
| BCS_Print_ActionList of action list 
| BCS_Print_TriggerList of trigger list 

let trigger_to_arg_list t ids = 
  if is_concat_string ids then 
    let a,b = split6 t.t_3 in
    [ (Arg_Integer,(Act_Integer t.t_1)) ;
      (Arg_Integer,(Act_Integer t.t_2)) ;
      (Arg_String,(Act_String a));
      (Arg_String,(Act_String b));
      (Arg_String,(Act_String t.t_4)); 
      (Arg_Object,(Act_Object t.t_5)) ; ]
  else 
    [ (Arg_Integer,(Act_Integer t.t_1)) ;
      (Arg_Integer,(Act_Integer t.t_2)) ;
      (Arg_String,(Act_String t.t_3)) ;
      (Arg_String,(Act_String t.t_4)) ;
      (Arg_Object,(Act_Object t.t_5)) ; ]

let action_to_arg_list a ids =
  if is_concat_string ids then
    let aa,b = split6 a.a_8 in
    [ (Arg_Object,(Act_Object a.a_2)) ;
      (Arg_Object,(Act_Object a.a_3)) ;
      (* (Arg_Object,(Act_Object a.a_1)) ; *)
      (Arg_Integer,(Act_Integer a.a_4)) ;
      (Arg_Point,(Act_Point a.a_5)) ;
      (Arg_Integer,(Act_Integer a.a_6)) ;
      (Arg_Integer,(Act_Integer a.a_7)) ;
      (Arg_String,(Act_String aa)) ;
      (Arg_String,(Act_String b)) ;
      (Arg_String,(Act_String a.a_9)) ; ]
  else
    [ (Arg_Object,(Act_Object a.a_2)) ;
      (Arg_Object,(Act_Object a.a_3)) ;
      (* (Arg_Object,(Act_Object a.a_1)) ; *)
      (Arg_Integer,(Act_Integer a.a_4)) ;
      (Arg_Point,(Act_Point a.a_5)) ;
      (Arg_Integer,(Act_Integer a.a_6)) ;
      (Arg_Integer,(Act_Integer a.a_7)) ;
      (Arg_String,(Act_String a.a_8)) ;
      (Arg_String,(Act_String a.a_9)) ; ]

let get_first_of_type argl tau =
  let res = List.assoc tau !argl in
  argl := List.remove_assoc tau !argl ;
  res

let formal_arg_is_strref arg = 
  let up = String.uppercase arg.arg_comment in
  up = "STRREF" || up = "STRINGREF"

let print_script_text game how what comments strhandle =
  let bcs_printf fmt = 
    let k result = match how with 
    | Save_BCS_OC(oc) -> output_string oc result
    | Save_BCS_Buffer(b) -> Buffer.add_string b result
    in 
    kprintf k fmt 
  in 
  let rec print_cr (tl,rl) =
    bcs_printf "IF\n" ;
    let or_count = ref 0 in 
    List.iter (fun t -> 
      let indent = 2 + if !or_count > 0 then (decr or_count ; 2) else 0 in 
      bcs_printf "%.*s" indent " " ;
      or_count := !or_count + (print_trigger t) ;
    ) tl ;
    bcs_printf "THEN\n" ;
    List.iter (fun (w,al) -> 
      bcs_printf "  RESPONSE #%d\n" w ;
      List.iter (fun a -> 
        bcs_printf "    " ;
        print_action a ;
      ) al 
    ) rl ;
    bcs_printf "END\n\n" 
  and print_obj obj =
    (* if obj.o_name <> "" then
      bcs_printf "\"%s\"" obj.o_name
    else *) begin

      let ilist = [ obj.o_unknown4 ; obj.o_unknown3 ; obj.o_unknown2 ;
        obj.o_unknown1 ; obj.o_identifiers ; ] 
      in 

      let dotlst = if game.Load.is_iwd2 then [
        obj.o_ea , "EA" ;
        obj.o_general , "GENERAL" ;
        obj.o_race , "RACE" ;
        obj.o_subrace , "SUBRACE" ;
        obj.o_class , "CLASS" ;
        obj.o_specific , "SPECIFIC" ;
        obj.o_gender , "GENDER" ;
        obj.o_alignment , "ALIGNMNT" 
      ] else [ 
        obj.o_ea , "EA" ;
        obj.o_general , "GENERAL" ;
        obj.o_race , "RACE" ;
        obj.o_class , "CLASS" ;
        obj.o_specific , "SPECIFIC" ;
        obj.o_gender , "GENDER" ;
        obj.o_alignment , "ALIGN" ] in

      let rec all_zeros lst = match lst with
      | [] -> true
      | (v,s) :: tl -> if v = Int32.zero then all_zeros tl else false
      in 

      let zero_dotlst = all_zeros dotlst in 

      let rec proc_dot lst = match lst with
      | [] -> ()
      | (v,s) :: tl -> 
        let str = if v = Int32.zero then "0" else sym_of_int game s v in
        if all_zeros tl then 
          bcs_printf "%s" str 
        else begin
          bcs_printf "%s." str ;
          proc_dot tl 
        end
      in 

      let rec proc lst lp rp anything = match lst with
      | [] -> 
          begin
            if all_zeros dotlst then begin
              (if not anything && obj.o_name = "" then bcs_printf "[ANYONE]"
               else if obj.o_name <> "" then bcs_printf "%s\"%s\"%s" lp obj.o_name rp)
            end else begin
              bcs_printf"%s[" lp ;
              proc_dot dotlst ;
              bcs_printf "]%s" rp ;
            end
          end 
      | hd :: tl ->
            if hd = Int32.zero then
              proc tl lp rp anything
            else begin
              let str = sym_of_int game "OBJECT" hd in
              bcs_printf "%s%s" lp str ;
              proc tl "(" ")" true ;
              bcs_printf "%s" rp ;
            end
      in 
      proc ilist "" "" false ; 
    end
  and print_arg_list argl ids =
    let commas = ref ((List.length ids.i_args) - 1) in 
    List.iter (fun a ->
      begin
      match get_first_of_type argl a.arg_kind with
      | (Act_Object obj) -> print_obj obj 
      | (Act_String s) -> bcs_printf "\"%s\"" s 
      | (Act_Integer i) -> begin
          if a.arg_file = "" then begin
            match strhandle with
              Some(f) when formal_arg_is_strref a -> begin
                let str = f i in
                bcs_printf "%s" str 
                end 
            | _ -> bcs_printf "%ld" i 
          end else begin
              let str = sym_of_int game a.arg_file i in
              bcs_printf "%s" str 
          end
          end
      | (Act_Point (x,y)) -> bcs_printf "[%ld.%ld]" x y
      end ;
      if !commas > 0 then begin
        bcs_printf "," ;
        decr commas 
      end 
    ) ids.i_args ;
  and print_action a =
    if non_empty_object a.a_1 then begin
      bcs_printf "ActionOverride(" ;
      print_obj a.a_1 ;
      bcs_printf "," ;
    end ; 
    let ids = best_ids_of_action game a in 
    bcs_printf "%s(" ids.i_name ;
    let argl = ref (action_to_arg_list a ids) in 
    print_arg_list argl ids ; 
    bcs_printf ")" ;
    if non_empty_object a.a_1 then begin bcs_printf ")" end ;
    if comments then begin 
      (* look for strref / stringref parameters *) 
      let contains_strref = List.exists formal_arg_is_strref ids.i_args in 
      if contains_strref then begin
        bcs_printf "  // %s" 
          (Tlk.one_line_pretty_print game.Load.dialog (Int32.to_int a.a_4))
      end ; 
      match ids.i_name with
      | "GiveItemCreate"
      | "GiveItem"
      | "CreateItem" 
      | "DestroyItem"
      | "UseItem"
      -> bcs_printf "  // %s" (name_of_res a.a_8 "ITM" 0xc)
      | "CreateCreature"
      | "CutSceneId"
      | "CreateCreatureOffScreen"
      | "CreateCreatureObject"
      | "CreateCreatureObjectDoor"
      | "CreateCreatureObjectOffScreen"
      -> bcs_printf "  // %s" (name_of_res a.a_8 "CRE" 0xc)
      | "AddSpecialAbility"
      | "ForceSpellRES"
      | "ReallyForceSpellRES"
      | "SpellRES"
      | "ApplySpellRES" 
      -> bcs_printf "  // %s" (name_of_res a.a_8 "SPL" 0x8)
      | "MarkSpellAndObject"
      -> 
      for i = 0 to (String.length a.a_8 / 4) - 1 do
        let str = 
          try
            let sub = String.sub a.a_8 (i*4) 4 in 
            let ids = ids_of_int game "SPELL" (Int32.of_string sub) in 
            ids.i_name 
          with _ -> "???" 
        in 
        bcs_printf "\n      // %s" str 
      done

      | _ -> () 
    end ;
    bcs_printf "\n" 
  and print_trigger t =  
    if (t.negated) then bcs_printf "!" ;
    let ids = best_ids_of_trigger game t in 
    bcs_printf "%s(" ids.i_name ; 
    let argl = ref (trigger_to_arg_list t ids) in 
    print_arg_list argl ids ; 
    bcs_printf ")" ;
    if comments then begin 
      match ids.i_name with
      | "PartyHasItem"
      | "HasItem"
      -> bcs_printf "  // %s" (name_of_res t.t_3 "ITM" 0xc)
      | "Dead"
      -> bcs_printf "  // %s" (name_of_res t.t_3 "CRE" 0xc)
      | _ -> 
        if t.t_5.o_name <> "" then 
          bcs_printf "  // %s" (name_of_res t.t_5.o_name "CRE" 0xc)
    end ; 
    bcs_printf "\n" ;
    if String.uppercase ids.i_name = "OR" then 
      (Int32.to_int t.t_1)
    else 0
  and name_of_res r ext offset =
    try 
      Load.skip_next_load_error := true ; 
      let buff,path = Load.load_resource game true r ext in
      Load.skip_next_load_error := false ; 
      if String.length buff >= offset + 4 then 
        (Tlk.one_line_pretty_print game.Load.dialog 
          ((int_of_str_off buff offset )))
      else r
    with _ -> r
  in 
  match what with 
  | BCS_Print_Script(s) -> List.iter print_cr s
  | BCS_Print_ActionList(al) -> List.iter print_action al 
  | BCS_Print_TriggerList(tl) -> List.iter 
      (fun t -> ignore (print_trigger t)) tl
