type token =
    QUESTION
  | COLON
  | ACTION_IF
  | ADD_KIT
  | ADD_MUSIC
  | ADD_STATE_TRIGGER
  | ADD_TRANS_TRIGGER
  | ALLOW_MISSING
  | ALWAYS
  | AND
  | OR
  | NOT
  | ELSE
  | LPAREN
  | RPAREN
  | APPEND
  | APPENDI
  | APPEND_COL
  | ASK_EVERY_COMPONENT
  | AT_EXIT
  | AT_INTERACTIVE_EXIT
  | AT_INTERACTIVE_UNINSTALL
  | SET_2DA_ENTRY
  | AT_UNINSTALL
  | AUTHOR
  | AUTO_TRA
  | BACKUP
  | BEGIN
  | CHAIN2
  | CHAIN3
  | COMPILE
  | COPY
  | GLOB
  | COPY_EXISTING
  | COPY_EXISTING_REGEXP
  | COPY_RANDOM
  | COPY_TRANS
  | DELETE_BYTES
  | DO
  | END
  | EQUALS
  | EQUALSEQUALS
  | EXIT
  | EXTEND_BOTTOM
  | EXTEND_BOTTOM_REGEXP
  | EXTEND_TOP
  | EXTEND_TOP_REGEXP
  | EXTERN
  | FAIL
  | FILE_CONTAINS
  | FILE_EXISTS
  | FILE_EXISTS_IN_GAME
  | FILE_SIZE
  | FLAGS
  | FORBID_FILE
  | GOTO
  | GT
  | GTE
  | LT
  | LTE
  | IF
  | SET
  | IF_EVAL
  | IF_SIZE_IS
  | INSERT_BYTES
  | INTERJECT
  | INTERJECT_COPY_TRANS
  | JOURNAL
  | SOLVED_JOURNAL
  | UNSOLVED_JOURNAL
  | LANGUAGE
  | MENU_STYLE
  | MKDIR
  | PLUS
  | MINUS
  | TIMES
  | DIVIDE
  | PRINT
  | RANDOM_SEED
  | READ_BYTE
  | READ_LONG
  | READ_SHORT
  | REPLACE
  | REPLACE_ACTION_TEXT
  | REPLACE_ACTION_TEXT_PROCESS
  | REPLACE_BCS_BLOCK
  | REPLACE_SAY
  | REPLACE_STATE_TRIGGER
  | REPLACE_TEXTUALLY
  | REPLACE_TRIGGER_TEXT
  | REPLY
  | REQUIRE_FILE
  | SAY
  | SET_WEIGHT
  | STRING of (string)
  | STRING_SET
  | THEN
  | UNINSTALL
  | UNLESS
  | USING
  | WEIGHT
  | WRITE_ASCII
  | WRITE_BYTE
  | WRITE_LONG
  | WRITE_SHORT
  | WHILE
  | EOF
  | SOUND of (string)
  | STRING_REF of (int)
  | TRANS_REF of (int)
  | FORCED_STRING_REF of (int)

open Parsing
# 2 "src/dparser.mly"
open Util
open Load

(*** input handle ***)

let current_unit = ref None

(*** Error handling ***)
let parse_error = Util.parse_error 

let get_current_unit () = match !current_unit with
  Some(s) -> s
| None -> parse_error "No current unit (use BEGIN)" 

type trans_opt_type =
    Trans_Reply of Dlg.tlk_string option
  | Trans_Do of string option
  | Trans_Journal of (Dlg.journal_type * Dlg.tlk_string) option
  | Trans_Flags of int

let extra_actions = ref [] 

let verify_trigger_list s =
  let con = the_context () in 
  let filename = (Printf.sprintf "trigger list near line %d, column %d of %s"
    con.line con.col con.filename) in 
  let lexbuf = lex_init_from_internal_string filename s in 
  try 
    let res = Bafparser.trigger_list Baflexer.initial lexbuf in 
    pop_context () ;
    let buff = Buffer.create (String.length s) in 
    Bcs.print_script_text (the_game()) (Bcs.Save_BCS_Buffer(buff))
        (Bcs.BCS_Print_TriggerList(res)) false None ;
    Buffer.contents buff 
  with e -> 
    pop_context () ; s 

let verify_action_list s =
  let con = the_context () in 
  let filename = (Printf.sprintf "action list near line %d, column %d of %s"
    con.line con.col con.filename) in 
  let lexbuf = lex_init_from_internal_string filename s in 
  try 
    let res = Bafparser.action_list Baflexer.initial lexbuf in 
    pop_context () ;
    let buff = Buffer.create (String.length s) in 
    Bcs.print_script_text (the_game()) (Bcs.Save_BCS_Buffer(buff))
        (Bcs.BCS_Print_ActionList(res)) false None ;
    Buffer.contents buff 
  with e -> pop_context () ; s

(* Line 55, file src/dparser.ml *)
let yytransl_const = [|
  257 (* QUESTION *);
  258 (* COLON *);
  259 (* ACTION_IF *);
  260 (* ADD_KIT *);
  261 (* ADD_MUSIC *);
  262 (* ADD_STATE_TRIGGER *);
  263 (* ADD_TRANS_TRIGGER *);
  264 (* ALLOW_MISSING *);
  265 (* ALWAYS *);
  266 (* AND *);
  267 (* OR *);
  268 (* NOT *);
  269 (* ELSE *);
  270 (* LPAREN *);
  271 (* RPAREN *);
  272 (* APPEND *);
  273 (* APPENDI *);
  274 (* APPEND_COL *);
  275 (* ASK_EVERY_COMPONENT *);
  276 (* AT_EXIT *);
  277 (* AT_INTERACTIVE_EXIT *);
  278 (* AT_INTERACTIVE_UNINSTALL *);
  279 (* SET_2DA_ENTRY *);
  280 (* AT_UNINSTALL *);
  281 (* AUTHOR *);
  282 (* AUTO_TRA *);
  283 (* BACKUP *);
  284 (* BEGIN *);
  285 (* CHAIN2 *);
  286 (* CHAIN3 *);
  287 (* COMPILE *);
  288 (* COPY *);
  289 (* GLOB *);
  290 (* COPY_EXISTING *);
  291 (* COPY_EXISTING_REGEXP *);
  292 (* COPY_RANDOM *);
  293 (* COPY_TRANS *);
  294 (* DELETE_BYTES *);
  295 (* DO *);
  296 (* END *);
  297 (* EQUALS *);
  298 (* EQUALSEQUALS *);
  299 (* EXIT *);
  300 (* EXTEND_BOTTOM *);
  301 (* EXTEND_BOTTOM_REGEXP *);
  302 (* EXTEND_TOP *);
  303 (* EXTEND_TOP_REGEXP *);
  304 (* EXTERN *);
  305 (* FAIL *);
  306 (* FILE_CONTAINS *);
  307 (* FILE_EXISTS *);
  308 (* FILE_EXISTS_IN_GAME *);
  309 (* FILE_SIZE *);
  310 (* FLAGS *);
  311 (* FORBID_FILE *);
  312 (* GOTO *);
  313 (* GT *);
  314 (* GTE *);
  315 (* LT *);
  316 (* LTE *);
  317 (* IF *);
  318 (* SET *);
  319 (* IF_EVAL *);
  320 (* IF_SIZE_IS *);
  321 (* INSERT_BYTES *);
  322 (* INTERJECT *);
  323 (* INTERJECT_COPY_TRANS *);
  324 (* JOURNAL *);
  325 (* SOLVED_JOURNAL *);
  326 (* UNSOLVED_JOURNAL *);
  327 (* LANGUAGE *);
  328 (* MENU_STYLE *);
  329 (* MKDIR *);
  330 (* PLUS *);
  331 (* MINUS *);
  332 (* TIMES *);
  333 (* DIVIDE *);
  334 (* PRINT *);
  335 (* RANDOM_SEED *);
  336 (* READ_BYTE *);
  337 (* READ_LONG *);
  338 (* READ_SHORT *);
  339 (* REPLACE *);
  340 (* REPLACE_ACTION_TEXT *);
  341 (* REPLACE_ACTION_TEXT_PROCESS *);
  342 (* REPLACE_BCS_BLOCK *);
  343 (* REPLACE_SAY *);
  344 (* REPLACE_STATE_TRIGGER *);
  345 (* REPLACE_TEXTUALLY *);
  346 (* REPLACE_TRIGGER_TEXT *);
  347 (* REPLY *);
  348 (* REQUIRE_FILE *);
  349 (* SAY *);
  350 (* SET_WEIGHT *);
  352 (* STRING_SET *);
  353 (* THEN *);
  354 (* UNINSTALL *);
  355 (* UNLESS *);
  356 (* USING *);
  357 (* WEIGHT *);
  358 (* WRITE_ASCII *);
  359 (* WRITE_BYTE *);
  360 (* WRITE_LONG *);
  361 (* WRITE_SHORT *);
  362 (* WHILE *);
    0 (* EOF *);
    0|]

let yytransl_block = [|
  351 (* STRING *);
  363 (* SOUND *);
  364 (* STRING_REF *);
  365 (* TRANS_REF *);
  366 (* FORCED_STRING_REF *);
    0|]

let yylhs = "\255\255\
\001\000\015\000\015\000\006\000\006\000\007\000\009\000\009\000\
\041\000\041\000\008\000\014\000\013\000\012\000\043\000\010\000\
\011\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
\023\000\023\000\047\000\047\000\046\000\046\000\046\000\046\000\
\046\000\044\000\044\000\024\000\024\000\045\000\045\000\017\000\
\017\000\021\000\021\000\048\000\048\000\049\000\049\000\050\000\
\018\000\018\000\018\000\042\000\022\000\022\000\022\000\026\000\
\026\000\019\000\019\000\029\000\029\000\029\000\029\000\029\000\
\029\000\030\000\030\000\051\000\052\000\052\000\020\000\020\000\
\020\000\028\000\028\000\028\000\028\000\025\000\025\000\025\000\
\025\000\025\000\027\000\027\000\002\000\002\000\003\000\032\000\
\032\000\032\000\032\000\032\000\053\000\053\000\040\000\040\000\
\031\000\031\000\033\000\033\000\034\000\034\000\054\000\054\000\
\055\000\055\000\056\000\056\000\035\000\035\000\035\000\035\000\
\035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
\035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
\035\000\035\000\035\000\035\000\035\000\035\000\035\000\057\000\
\057\000\057\000\057\000\057\000\057\000\057\000\057\000\039\000\
\039\000\036\000\036\000\036\000\036\000\036\000\037\000\037\000\
\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
\058\000\058\000\058\000\058\000\058\000\058\000\058\000\038\000\
\038\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
\059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
\059\000\004\000\005\000\005\000\000\000\000\000\000\000\000\000"

let yylen = "\002\000\
\001\000\000\000\002\000\002\000\003\000\002\000\004\000\004\000\
\000\000\001\000\002\000\005\000\005\000\005\000\004\000\004\000\
\004\000\002\000\003\000\003\000\003\000\004\000\001\000\001\000\
\001\000\004\000\004\000\005\000\003\000\004\000\004\000\004\000\
\003\000\005\000\000\000\002\000\003\000\003\000\003\000\001\000\
\002\000\000\000\004\000\000\000\003\000\000\000\004\000\000\000\
\002\000\000\000\002\000\000\000\001\000\000\000\001\000\001\000\
\010\000\003\000\007\000\002\000\001\000\003\000\003\000\001\000\
\003\000\000\000\002\000\002\000\002\000\002\000\002\000\002\000\
\002\000\000\000\002\000\001\000\001\000\000\000\005\000\006\000\
\003\000\002\000\002\000\003\000\001\000\002\000\004\000\001\000\
\001\000\002\000\000\000\001\000\000\000\004\000\008\000\000\000\
\003\000\003\000\002\000\004\000\000\000\002\000\000\000\002\000\
\000\000\005\000\000\000\004\000\000\000\002\000\000\000\002\000\
\000\000\001\000\000\000\001\000\006\000\005\000\005\000\003\000\
\002\000\003\000\002\000\003\000\003\000\004\000\004\000\005\000\
\005\000\005\000\005\000\002\000\002\000\002\000\002\000\003\000\
\003\000\003\000\021\000\002\000\002\000\006\000\010\000\003\000\
\003\000\003\000\002\000\002\000\002\000\003\000\003\000\000\000\
\002\000\000\000\003\000\003\000\003\000\003\000\000\000\003\000\
\001\000\003\000\003\000\003\000\003\000\003\000\003\000\002\000\
\003\000\003\000\003\000\003\000\003\000\003\000\005\000\000\000\
\002\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
\003\000\003\000\003\000\003\000\003\000\003\000\004\000\005\000\
\005\000\001\000\000\000\004\000\002\000\002\000\002\000\002\000"

let yydefred = "\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\197\000\000\000\000\000\
\000\000\000\000\000\000\000\000\024\000\025\000\023\000\001\000\
\000\000\000\000\000\000\198\000\000\000\199\000\000\000\200\000\
\194\000\000\000\000\000\006\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\011\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\018\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\003\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\005\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\060\000\000\000\000\000\049\000\000\000\019\000\
\021\000\000\000\076\000\000\000\077\000\000\000\020\000\067\000\
\000\000\000\000\029\000\000\000\000\000\000\000\088\000\089\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\051\000\
\000\000\015\000\153\000\010\000\008\000\007\000\016\000\017\000\
\031\000\032\000\022\000\000\000\030\000\026\000\000\000\056\000\
\000\000\058\000\081\000\053\000\000\000\000\000\000\000\000\000\
\045\000\000\000\000\000\040\000\000\000\027\000\092\000\000\000\
\090\000\000\000\000\000\094\000\000\000\196\000\014\000\013\000\
\043\000\012\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\047\000\028\000\
\000\000\000\000\041\000\000\000\000\000\036\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\055\000\
\000\000\069\000\073\000\070\000\071\000\072\000\068\000\075\000\
\085\000\000\000\000\000\000\000\079\000\000\000\039\000\037\000\
\038\000\087\000\034\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\099\000\
\000\000\000\000\000\000\000\000\000\000\062\000\063\000\000\000\
\000\000\082\000\083\000\080\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\132\000\
\133\000\134\000\135\000\000\000\114\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\140\000\000\000\123\000\
\141\000\121\000\000\000\000\000\000\000\000\000\110\000\097\000\
\098\000\000\000\102\000\000\000\000\000\059\000\000\000\084\000\
\147\000\000\000\000\000\148\000\149\000\000\000\000\000\000\000\
\000\000\000\000\136\000\000\000\000\000\000\000\122\000\116\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\120\000\000\000\000\000\000\000\
\000\000\000\000\125\000\124\000\137\000\138\000\100\000\104\000\
\000\000\000\000\095\000\000\000\000\000\144\000\151\000\150\000\
\000\000\146\000\000\000\000\000\000\000\000\000\000\000\000\000\
\126\000\127\000\112\000\000\000\000\000\000\000\000\000\000\000\
\000\000\161\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\177\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\160\000\118\000\119\000\168\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\182\000\
\184\000\183\000\179\000\181\000\180\000\178\000\188\000\000\000\
\000\000\000\000\000\000\129\000\131\000\128\000\130\000\000\000\
\000\000\065\000\057\000\000\000\000\000\155\000\158\000\157\000\
\156\000\117\000\162\000\000\000\000\000\000\000\167\000\171\000\
\172\000\173\000\174\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\106\000\108\000\000\000\000\000\000\000\193\000\
\192\000\000\000\000\000\000\000\000\000\000\000\143\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\139\000"

let yydgoto = "\005\000\
\022\000\036\000\038\000\040\000\041\000\023\000\024\000\025\000\
\026\000\027\000\028\000\029\000\030\000\031\000\032\000\033\000\
\062\000\063\000\070\000\071\000\086\000\171\000\077\000\078\000\
\172\000\101\001\160\000\213\000\180\000\181\000\045\001\196\000\
\099\001\246\000\247\000\113\001\067\001\085\001\089\000\043\001\
\133\000\064\000\034\000\047\000\073\000\158\000\163\000\149\000\
\201\000\145\000\108\000\110\000\252\000\063\001\022\001\065\001\
\011\001\123\001\086\001"

let yysindex = "\193\000\
\144\013\154\254\006\255\182\254\000\000\203\254\211\254\216\254\
\225\254\007\255\232\254\234\254\236\254\242\254\012\255\041\255\
\051\255\062\255\064\255\067\255\082\255\000\000\019\255\019\255\
\019\255\124\255\110\255\110\255\000\000\000\000\000\000\000\000\
\144\013\119\255\147\255\000\000\095\255\000\000\194\254\000\000\
\000\000\115\255\121\255\000\000\129\255\125\255\133\255\134\255\
\134\255\141\255\159\255\000\000\160\255\167\255\168\255\174\255\
\175\255\176\255\177\255\219\255\125\255\000\000\019\255\019\255\
\022\000\024\000\227\255\228\255\234\255\042\000\124\255\245\255\
\045\000\055\000\000\000\001\000\110\255\165\255\165\255\085\000\
\007\000\017\000\018\000\000\000\008\000\027\000\037\000\134\255\
\011\000\011\000\041\000\047\000\048\000\053\000\165\255\057\000\
\059\000\060\000\000\000\066\000\077\000\000\000\134\000\000\000\
\000\000\081\000\000\000\080\000\000\000\111\000\000\000\000\000\
\119\255\092\000\000\000\091\000\086\255\082\000\000\000\000\000\
\165\255\152\000\154\254\097\000\182\254\134\255\134\255\000\000\
\096\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\134\255\000\000\000\000\165\255\000\000\
\080\000\000\000\000\000\000\000\247\254\165\255\110\255\099\000\
\000\000\100\000\084\255\000\000\106\000\000\000\000\000\107\000\
\000\000\108\000\163\000\000\000\025\255\000\000\000\000\000\000\
\000\000\000\000\165\000\241\254\191\000\127\000\128\000\165\255\
\165\255\165\255\165\255\247\254\156\255\247\254\000\000\000\000\
\130\000\131\000\000\000\137\000\082\000\000\000\119\255\018\013\
\025\255\141\000\147\000\221\000\155\000\165\255\165\255\000\000\
\156\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\157\000\158\000\162\000\000\000\156\255\000\000\000\000\
\000\000\000\000\000\000\040\255\166\000\168\000\174\000\175\000\
\176\000\178\000\179\000\182\000\134\255\207\000\207\000\207\000\
\134\255\188\000\190\000\192\000\199\000\165\255\205\000\134\255\
\165\255\206\000\209\000\215\000\216\000\246\000\018\013\000\000\
\025\255\025\255\217\000\231\000\218\000\000\000\000\000\223\000\
\219\000\000\000\000\000\000\000\040\255\040\255\233\000\234\000\
\236\000\237\000\005\255\239\000\240\000\243\000\250\000\000\000\
\000\000\000\000\000\000\241\000\000\000\062\001\008\001\008\001\
\171\013\009\001\010\001\011\001\015\001\000\000\165\255\000\000\
\000\000\000\000\165\255\165\255\019\001\025\255\000\000\000\000\
\000\000\217\000\000\000\021\001\094\001\000\000\165\255\000\000\
\000\000\122\255\028\001\000\000\000\000\029\001\040\255\040\255\
\098\001\032\001\000\000\056\255\056\255\134\255\000\000\000\000\
\008\001\035\001\171\013\171\013\255\254\255\254\039\001\255\254\
\255\254\255\254\255\254\041\001\043\001\044\001\255\254\255\254\
\255\254\255\254\255\254\255\254\000\000\171\013\171\013\171\013\
\171\013\171\013\000\000\000\000\000\000\000\000\000\000\000\000\
\045\001\165\255\000\000\106\001\124\255\000\000\000\000\000\000\
\138\001\000\000\018\013\058\001\059\001\255\254\060\001\069\001\
\000\000\000\000\000\000\171\013\008\001\056\255\056\255\255\254\
\255\254\000\000\045\013\045\013\116\001\045\013\087\012\067\013\
\071\013\165\255\070\001\072\001\143\255\122\013\045\013\045\013\
\045\013\238\013\000\000\241\000\241\000\241\000\241\000\134\255\
\018\013\165\255\133\001\141\001\087\001\056\255\182\255\056\255\
\056\255\056\255\000\000\000\000\000\000\000\000\014\014\255\254\
\255\254\255\254\255\254\255\254\255\254\255\254\255\254\255\254\
\255\254\255\254\255\254\045\013\021\014\255\254\021\014\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\021\014\
\021\014\021\014\171\013\000\000\000\000\000\000\000\000\231\000\
\094\001\000\000\000\000\171\001\090\001\000\000\000\000\000\000\
\000\000\000\000\000\000\245\013\145\010\190\255\000\000\000\000\
\000\000\000\000\000\000\147\004\147\004\220\255\220\255\148\013\
\021\014\147\001\000\000\000\000\160\001\096\001\255\254\000\000\
\000\000\018\013\099\001\148\255\158\001\109\001\000\000\117\001\
\118\001\119\001\120\001\124\001\130\001\165\255\132\001\165\255\
\139\001\165\255\000\000"

let yyrindex = "\000\000\
\231\002\233\002\000\000\235\002\000\000\000\000\000\000\000\000\
\000\000\203\254\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\093\012\196\001\
\196\001\199\001\203\001\203\001\000\000\000\000\000\000\000\000\
\231\002\240\255\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\251\011\236\254\000\000\251\254\
\251\254\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\236\254\000\000\020\012\196\001\
\000\000\000\000\000\000\000\000\173\001\000\000\085\012\000\000\
\000\000\000\000\000\000\000\000\126\001\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\091\004\
\104\000\104\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\112\009\000\000\000\000\000\000\000\000\
\240\255\000\000\000\000\000\000\000\000\020\003\000\000\000\000\
\000\000\255\000\233\002\000\000\235\002\178\012\178\012\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\178\012\000\000\000\000\000\000\000\000\
\250\254\000\000\000\000\000\000\035\000\000\000\126\001\000\000\
\000\000\000\000\186\012\000\000\000\000\000\000\000\000\127\003\
\000\000\000\000\108\001\000\000\004\000\000\000\000\000\000\000\
\000\000\000\000\000\000\205\001\051\255\000\000\000\000\000\000\
\000\000\000\000\000\000\035\000\000\000\035\000\000\000\000\000\
\000\000\000\000\000\000\000\000\234\003\000\000\240\255\208\001\
\004\000\000\000\000\000\005\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\143\009\198\004\049\005\049\005\
\218\008\000\000\000\000\000\000\000\000\000\000\000\000\133\010\
\000\000\000\000\000\000\000\000\000\000\000\000\014\000\000\000\
\004\000\004\000\006\000\003\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\212\010\000\000\154\005\114\008\114\008\
\035\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\
\000\000\006\000\000\000\000\000\249\002\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\114\011\114\011\133\010\000\000\000\000\
\114\008\000\000\049\010\049\010\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\058\009\224\009\224\009\
\224\009\224\009\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\146\000\199\001\000\000\000\000\000\000\
\014\255\000\000\208\001\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\049\010\114\008\114\011\114\011\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\212\010\212\010\212\010\212\010\010\000\
\023\000\000\000\000\000\000\000\000\000\114\011\114\011\114\011\
\114\011\114\011\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\002\006\000\000\106\006\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\210\006\
\058\007\162\007\214\001\000\000\000\000\000\000\000\000\003\000\
\249\002\000\000\000\000\193\011\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\168\002\059\002\000\000\000\000\
\000\000\000\000\000\000\097\001\206\001\135\000\244\000\000\000\
\010\008\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\208\001\000\000\025\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000"

let yygindex = "\000\000\
\000\000\133\002\000\000\000\000\132\002\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\225\002\000\000\
\118\000\000\000\205\255\000\000\198\002\153\255\171\255\000\000\
\178\255\114\001\073\002\049\002\000\000\123\255\073\001\072\255\
\071\001\011\255\000\000\253\254\244\254\207\254\238\255\224\001\
\182\002\000\000\000\000\000\000\236\255\000\000\000\000\128\002\
\000\000\000\000\000\000\000\000\000\000\092\255\211\255\000\000\
\057\255\004\255\000\000"

let yytablesize = 3938
let yytable = "\122\000\
\123\000\039\001\105\000\096\000\101\000\103\000\035\000\074\000\
\248\000\152\000\120\001\068\001\121\001\109\000\055\001\056\001\
\139\000\118\001\119\001\112\000\039\000\052\000\109\000\145\000\
\175\000\198\000\199\000\151\000\145\000\174\000\090\000\152\000\
\037\000\192\000\152\000\059\000\139\001\140\001\141\001\142\001\
\143\001\042\000\161\000\193\000\175\000\081\000\208\000\060\000\
\214\000\043\000\194\000\005\001\116\001\006\001\044\000\152\000\
\117\000\114\001\176\000\177\000\178\000\049\001\050\001\045\000\
\040\001\041\001\154\001\046\000\152\000\131\000\048\000\182\000\
\049\000\124\001\050\000\126\001\127\001\128\001\129\001\061\000\
\051\000\179\000\133\001\134\001\135\001\136\001\137\001\138\001\
\052\000\007\001\008\001\009\001\010\001\122\001\254\000\255\000\
\195\000\204\000\205\000\206\000\207\000\057\001\152\000\187\000\
\155\001\219\000\052\000\167\000\168\000\095\001\145\000\105\001\
\106\001\151\001\156\001\157\001\109\001\148\001\110\001\111\001\
\067\000\170\000\154\000\158\001\159\001\155\000\172\001\173\001\
\156\000\175\001\183\000\055\001\056\001\157\000\165\000\053\000\
\102\001\218\001\184\001\185\001\186\001\065\000\066\000\160\001\
\068\000\054\000\198\001\199\001\200\001\201\001\202\001\072\000\
\161\001\162\001\112\001\193\001\055\000\069\000\056\000\030\001\
\067\000\057\000\033\001\204\001\205\001\206\001\207\001\208\001\
\209\001\210\001\211\001\212\001\213\001\214\001\215\001\216\001\
\058\000\217\001\186\000\076\000\102\000\103\000\160\001\163\001\
\068\000\023\001\024\001\079\000\163\001\080\000\160\001\161\001\
\162\001\001\000\002\000\003\000\004\000\069\000\209\000\164\001\
\165\001\166\001\167\001\210\000\164\001\165\001\166\001\167\001\
\091\001\082\000\020\001\211\000\092\001\093\001\025\001\083\000\
\168\001\169\001\170\001\171\001\160\001\032\001\163\001\084\000\
\100\001\085\000\228\001\087\000\088\000\212\000\163\001\188\001\
\189\001\190\001\191\001\091\000\229\001\118\000\164\001\165\001\
\166\001\167\001\109\001\166\000\110\001\111\001\164\001\165\001\
\166\001\167\001\119\000\120\000\121\000\092\000\093\000\168\001\
\169\001\170\001\171\001\118\000\163\001\094\000\095\000\168\001\
\169\001\170\001\171\001\096\000\096\000\097\000\098\000\099\000\
\119\000\120\000\121\000\145\001\164\001\165\001\166\001\167\001\
\112\001\175\000\175\000\175\000\175\000\175\000\105\000\096\000\
\101\000\103\000\175\000\175\000\175\000\152\000\175\000\175\000\
\175\000\109\000\175\000\115\001\175\000\175\000\175\000\175\000\
\175\000\147\001\109\000\179\001\175\000\109\000\182\001\175\000\
\175\000\100\000\175\000\175\000\175\000\104\000\175\000\105\000\
\175\000\106\000\107\000\100\001\175\000\175\000\175\000\175\000\
\109\000\175\000\096\000\101\000\103\000\074\000\044\000\175\000\
\152\000\111\000\074\000\113\000\114\000\175\000\175\000\175\000\
\175\000\175\000\074\000\044\000\044\000\044\000\115\000\116\000\
\163\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
\175\000\175\000\175\000\175\000\074\000\124\000\175\000\126\000\
\127\000\175\000\125\000\128\000\175\000\175\000\132\000\175\000\
\175\000\129\000\175\000\175\000\175\000\192\001\175\000\175\000\
\175\000\175\000\175\000\130\000\175\000\175\000\175\000\135\000\
\165\000\165\000\165\000\165\000\009\000\136\000\137\000\009\000\
\165\000\165\000\165\000\138\000\165\000\165\000\165\000\140\000\
\165\000\141\000\165\000\165\000\165\000\165\000\165\000\239\001\
\143\000\241\001\165\000\243\001\009\000\165\000\165\000\142\000\
\165\000\165\000\165\000\144\000\165\000\146\000\165\000\147\000\
\148\000\009\000\165\000\165\000\165\000\165\000\064\000\165\000\
\150\000\064\000\152\000\153\000\159\000\165\000\162\000\165\000\
\169\000\184\000\185\000\165\000\165\000\165\000\165\000\165\000\
\188\000\189\000\190\000\191\000\197\000\164\000\064\000\165\000\
\165\000\165\000\165\000\165\000\165\000\165\000\165\000\165\000\
\165\000\165\000\200\000\064\000\165\000\202\000\203\000\165\000\
\215\000\216\000\165\000\165\000\251\000\165\000\165\000\217\000\
\165\000\165\000\165\000\249\000\165\000\165\000\165\000\165\000\
\165\000\250\000\165\000\165\000\165\000\166\000\166\000\166\000\
\166\000\253\000\000\001\001\001\002\001\166\000\166\000\166\000\
\003\001\166\000\166\000\166\000\012\001\166\000\013\001\166\000\
\166\000\166\000\166\000\166\000\014\001\015\001\016\001\166\000\
\017\001\018\001\166\000\166\000\019\001\166\000\166\000\166\000\
\021\001\166\000\026\001\166\000\027\001\038\001\028\001\166\000\
\166\000\166\000\166\000\035\000\166\000\029\001\035\000\035\000\
\035\000\035\000\166\000\031\001\034\001\044\001\035\000\035\001\
\166\000\166\000\166\000\166\000\166\000\036\001\037\001\042\001\
\046\001\048\001\170\000\047\001\166\000\166\000\166\000\166\000\
\166\000\166\000\166\000\166\000\166\000\166\000\166\000\051\001\
\052\001\166\000\053\001\054\001\166\000\058\001\059\001\166\000\
\166\000\060\001\166\000\166\000\062\001\166\000\166\000\166\000\
\061\001\166\000\166\000\166\000\166\000\166\000\064\001\166\000\
\166\000\166\000\163\000\163\000\163\000\163\000\066\001\087\001\
\088\001\089\001\163\000\163\000\163\000\090\001\163\000\163\000\
\163\000\094\001\163\000\097\001\163\000\163\000\163\000\163\000\
\163\000\098\001\103\001\104\001\163\000\107\001\108\001\163\000\
\163\000\117\001\163\000\163\000\163\000\125\001\163\000\130\001\
\163\000\131\001\132\001\144\001\163\000\163\000\163\000\163\000\
\033\000\163\000\146\001\033\000\056\001\033\000\033\000\163\000\
\149\001\150\001\152\001\033\000\174\001\163\000\163\000\163\000\
\163\000\163\000\046\000\153\001\180\001\046\000\181\001\169\000\
\046\000\163\000\163\000\163\000\195\001\046\000\163\000\163\000\
\163\000\163\000\163\000\163\000\196\001\197\001\163\000\221\001\
\222\001\163\000\225\001\226\001\163\000\163\000\227\001\163\000\
\163\000\230\001\163\000\163\000\163\000\231\001\163\000\163\000\
\163\000\163\000\163\000\232\001\163\000\163\000\163\000\164\000\
\164\000\164\000\164\000\233\001\234\001\235\001\236\001\164\000\
\164\000\164\000\237\001\164\000\164\000\164\000\238\001\164\000\
\240\001\164\000\164\000\164\000\164\000\164\000\002\000\242\001\
\093\000\164\000\195\000\048\000\164\000\164\000\066\000\164\000\
\164\000\164\000\046\000\164\000\061\000\164\000\078\000\109\000\
\107\000\164\000\164\000\164\000\164\000\176\000\164\000\164\000\
\166\000\075\000\101\000\194\001\164\000\218\000\004\001\220\001\
\219\001\096\001\164\000\164\000\164\000\164\000\164\000\134\000\
\173\000\000\000\000\000\091\000\000\000\000\000\164\000\164\000\
\164\000\000\000\000\000\164\000\164\000\164\000\164\000\164\000\
\164\000\000\000\000\000\164\000\000\000\000\000\164\000\000\000\
\000\000\164\000\164\000\000\000\164\000\164\000\000\000\164\000\
\164\000\164\000\000\000\164\000\164\000\164\000\164\000\164\000\
\000\000\164\000\164\000\164\000\170\000\170\000\170\000\170\000\
\000\000\000\000\000\000\000\000\170\000\170\000\170\000\000\000\
\170\000\170\000\170\000\000\000\170\000\000\000\170\000\170\000\
\170\000\170\000\170\000\000\000\000\000\000\000\170\000\000\000\
\000\000\170\000\170\000\000\000\170\000\170\000\170\000\000\000\
\170\000\000\000\170\000\000\000\000\000\000\000\170\000\170\000\
\170\000\170\000\000\000\170\000\000\000\000\000\000\000\000\000\
\000\000\170\000\000\000\000\000\000\000\000\000\000\000\170\000\
\170\000\170\000\170\000\170\000\000\000\000\000\086\000\000\000\
\000\000\000\000\000\000\170\000\000\000\000\000\000\000\000\000\
\170\000\170\000\170\000\170\000\170\000\170\000\000\000\000\000\
\170\000\000\000\000\000\170\000\000\000\000\000\170\000\170\000\
\000\000\170\000\170\000\000\000\170\000\170\000\170\000\000\000\
\170\000\170\000\170\000\170\000\170\000\000\000\170\000\170\000\
\170\000\169\000\169\000\169\000\169\000\000\000\000\000\000\000\
\000\000\169\000\000\000\169\000\000\000\169\000\169\000\169\000\
\000\000\169\000\000\000\169\000\169\000\169\000\169\000\169\000\
\000\000\000\000\000\000\169\000\000\000\000\000\169\000\169\000\
\000\000\169\000\169\000\169\000\000\000\169\000\000\000\169\000\
\000\000\000\000\000\000\169\000\169\000\169\000\169\000\000\000\
\169\000\000\000\000\000\000\000\000\000\000\000\169\000\000\000\
\000\000\000\000\000\000\000\000\169\000\169\000\169\000\169\000\
\169\000\091\000\000\000\000\000\000\000\000\000\000\000\000\000\
\169\000\000\000\000\000\000\000\000\000\169\000\169\000\169\000\
\169\000\169\000\169\000\000\000\000\000\169\000\000\000\000\000\
\169\000\000\000\000\000\169\000\169\000\000\000\169\000\169\000\
\000\000\169\000\169\000\169\000\000\000\169\000\169\000\169\000\
\169\000\169\000\000\000\169\000\169\000\169\000\091\000\091\000\
\091\000\091\000\091\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\091\000\000\000\091\000\000\000\091\000\
\091\000\091\000\091\000\091\000\000\000\000\000\000\000\091\000\
\000\000\091\000\091\000\091\000\000\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\000\000\000\000\000\000\
\000\000\091\000\091\000\091\000\000\000\000\000\000\000\000\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\152\000\000\000\091\000\091\000\000\000\000\000\
\000\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\000\000\091\000\091\000\091\000\
\000\000\091\000\091\000\091\000\091\000\091\000\000\000\000\000\
\091\000\086\000\086\000\086\000\086\000\086\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\086\000\000\000\
\086\000\000\000\086\000\086\000\086\000\086\000\086\000\000\000\
\000\000\000\000\086\000\000\000\086\000\086\000\086\000\000\000\
\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
\000\000\000\000\000\000\000\000\086\000\086\000\086\000\000\000\
\000\000\000\000\000\000\086\000\086\000\086\000\086\000\086\000\
\086\000\086\000\086\000\086\000\086\000\113\000\000\000\086\000\
\086\000\000\000\000\000\000\000\086\000\086\000\086\000\086\000\
\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
\086\000\086\000\086\000\086\000\086\000\000\000\086\000\000\000\
\086\000\086\000\086\000\000\000\086\000\086\000\086\000\086\000\
\086\000\000\000\000\000\086\000\091\000\091\000\091\000\091\000\
\091\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\091\000\000\000\091\000\000\000\091\000\091\000\091\000\
\091\000\091\000\000\000\000\000\000\000\091\000\000\000\091\000\
\091\000\091\000\000\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\000\000\000\000\000\000\000\000\091\000\
\091\000\091\000\000\000\000\000\000\000\000\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\113\000\000\000\091\000\091\000\000\000\000\000\000\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
\000\000\091\000\000\000\091\000\091\000\091\000\000\000\091\000\
\091\000\091\000\091\000\091\000\000\000\000\000\091\000\000\000\
\000\000\000\000\000\000\000\000\000\000\152\000\152\000\152\000\
\152\000\152\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\152\000\000\000\152\000\000\000\152\000\152\000\
\152\000\152\000\152\000\000\000\000\000\000\000\152\000\000\000\
\152\000\152\000\152\000\000\000\152\000\152\000\152\000\152\000\
\152\000\000\000\152\000\000\000\000\000\000\000\152\000\152\000\
\152\000\152\000\000\000\152\000\000\000\000\000\000\000\000\000\
\000\000\152\000\000\000\160\001\000\000\000\000\000\000\152\000\
\152\000\115\000\000\000\152\000\152\000\152\000\000\000\000\000\
\000\000\152\000\000\000\152\000\152\000\000\000\000\000\000\000\
\152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
\152\000\152\000\152\000\152\000\152\000\000\000\152\000\152\000\
\152\000\000\000\152\000\163\001\152\000\000\000\152\000\000\000\
\152\000\152\000\152\000\152\000\152\000\000\000\152\000\000\000\
\113\000\113\000\113\000\164\001\165\001\166\001\167\001\000\000\
\000\000\000\000\000\000\000\000\000\000\113\000\000\000\113\000\
\000\000\113\000\113\000\113\000\113\000\113\000\170\001\171\001\
\000\000\113\000\000\000\000\000\113\000\113\000\113\000\113\000\
\113\000\113\000\000\000\113\000\000\000\113\000\000\000\000\000\
\000\000\113\000\113\000\113\000\113\000\000\000\113\000\000\000\
\000\000\000\000\000\000\000\000\113\000\000\000\000\000\000\000\
\000\000\190\000\113\000\113\000\113\000\113\000\113\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\113\000\000\000\
\000\000\000\000\000\000\113\000\113\000\113\000\113\000\113\000\
\113\000\000\000\000\000\113\000\000\000\000\000\113\000\000\000\
\000\000\113\000\113\000\000\000\113\000\113\000\000\000\113\000\
\113\000\000\000\000\000\113\000\113\000\113\000\113\000\113\000\
\000\000\000\000\000\000\113\000\113\000\113\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\113\000\000\000\113\000\000\000\113\000\113\000\113\000\113\000\
\113\000\000\000\000\000\000\000\113\000\000\000\000\000\113\000\
\113\000\000\000\113\000\113\000\113\000\000\000\113\000\000\000\
\113\000\000\000\000\000\000\000\113\000\113\000\113\000\113\000\
\000\000\113\000\000\000\000\000\000\000\000\000\000\000\113\000\
\000\000\189\000\000\000\000\000\000\000\113\000\113\000\113\000\
\113\000\113\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\113\000\000\000\000\000\000\000\000\000\113\000\113\000\
\113\000\113\000\113\000\113\000\000\000\000\000\113\000\000\000\
\000\000\113\000\000\000\000\000\113\000\113\000\000\000\113\000\
\113\000\000\000\113\000\113\000\000\000\000\000\113\000\113\000\
\113\000\113\000\113\000\000\000\115\000\115\000\115\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\115\000\000\000\115\000\000\000\115\000\115\000\115\000\
\115\000\115\000\000\000\000\000\000\000\115\000\000\000\000\000\
\115\000\115\000\000\000\115\000\115\000\115\000\000\000\115\000\
\000\000\115\000\000\000\000\000\000\000\115\000\115\000\115\000\
\115\000\000\000\115\000\000\000\000\000\000\000\000\000\000\000\
\115\000\185\000\000\000\000\000\000\000\000\000\115\000\115\000\
\115\000\115\000\115\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\115\000\000\000\000\000\000\000\000\000\115\000\
\115\000\115\000\115\000\115\000\115\000\000\000\000\000\115\000\
\000\000\000\000\115\000\000\000\000\000\115\000\115\000\000\000\
\115\000\115\000\000\000\115\000\115\000\000\000\000\000\115\000\
\115\000\115\000\115\000\115\000\190\000\190\000\190\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\190\000\000\000\190\000\000\000\190\000\190\000\190\000\
\190\000\190\000\000\000\000\000\000\000\190\000\000\000\000\000\
\190\000\190\000\000\000\190\000\190\000\190\000\000\000\190\000\
\000\000\190\000\000\000\000\000\000\000\190\000\190\000\190\000\
\190\000\000\000\190\000\000\000\000\000\000\000\000\000\000\000\
\190\000\187\000\000\000\000\000\000\000\000\000\190\000\190\000\
\190\000\190\000\190\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\190\000\000\000\000\000\000\000\000\000\190\000\
\190\000\190\000\190\000\190\000\190\000\000\000\000\000\190\000\
\000\000\000\000\190\000\000\000\000\000\190\000\190\000\000\000\
\000\000\190\000\000\000\190\000\190\000\190\000\000\000\190\000\
\190\000\190\000\190\000\190\000\189\000\189\000\189\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\189\000\000\000\189\000\000\000\189\000\189\000\189\000\
\189\000\189\000\000\000\000\000\000\000\189\000\000\000\000\000\
\189\000\189\000\000\000\189\000\189\000\189\000\000\000\189\000\
\000\000\189\000\000\000\000\000\000\000\189\000\189\000\189\000\
\189\000\000\000\189\000\000\000\000\000\000\000\000\000\000\000\
\189\000\186\000\000\000\000\000\000\000\000\000\189\000\189\000\
\189\000\189\000\189\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\189\000\000\000\000\000\000\000\000\000\189\000\
\189\000\189\000\189\000\189\000\189\000\000\000\000\000\189\000\
\000\000\000\000\189\000\000\000\000\000\189\000\189\000\000\000\
\000\000\189\000\000\000\189\000\189\000\189\000\000\000\189\000\
\189\000\189\000\189\000\189\000\185\000\185\000\185\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\185\000\000\000\185\000\000\000\185\000\185\000\185\000\
\185\000\185\000\000\000\000\000\000\000\185\000\000\000\000\000\
\185\000\185\000\000\000\185\000\185\000\185\000\000\000\185\000\
\000\000\185\000\000\000\000\000\000\000\185\000\185\000\185\000\
\185\000\000\000\185\000\000\000\000\000\000\000\000\000\000\000\
\185\000\191\000\000\000\000\000\000\000\000\000\185\000\185\000\
\185\000\185\000\185\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\185\000\000\000\000\000\000\000\000\000\185\000\
\185\000\185\000\185\000\185\000\185\000\000\000\000\000\185\000\
\000\000\000\000\185\000\000\000\000\000\185\000\185\000\000\000\
\000\000\185\000\000\000\185\000\185\000\185\000\000\000\185\000\
\185\000\185\000\185\000\185\000\187\000\187\000\187\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\187\000\000\000\187\000\000\000\187\000\187\000\187\000\
\187\000\187\000\000\000\000\000\000\000\187\000\000\000\000\000\
\187\000\187\000\000\000\187\000\187\000\187\000\000\000\187\000\
\000\000\187\000\000\000\000\000\000\000\187\000\187\000\187\000\
\187\000\000\000\187\000\000\000\000\000\000\000\000\000\000\000\
\187\000\159\000\000\000\000\000\000\000\000\000\187\000\187\000\
\187\000\187\000\187\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\187\000\000\000\000\000\000\000\000\000\187\000\
\187\000\187\000\187\000\187\000\187\000\000\000\000\000\187\000\
\000\000\000\000\187\000\000\000\000\000\187\000\187\000\000\000\
\000\000\187\000\000\000\187\000\187\000\187\000\000\000\187\000\
\187\000\187\000\187\000\187\000\186\000\186\000\186\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\186\000\000\000\186\000\000\000\186\000\186\000\186\000\
\186\000\186\000\000\000\000\000\000\000\186\000\000\000\000\000\
\186\000\186\000\000\000\186\000\186\000\186\000\000\000\186\000\
\000\000\186\000\000\000\000\000\000\000\186\000\186\000\186\000\
\186\000\000\000\186\000\000\000\000\000\000\000\000\000\000\000\
\186\000\152\000\000\000\000\000\000\000\000\000\186\000\186\000\
\186\000\186\000\186\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\186\000\000\000\000\000\000\000\000\000\186\000\
\186\000\186\000\186\000\186\000\186\000\000\000\000\000\186\000\
\000\000\000\000\186\000\000\000\000\000\186\000\186\000\000\000\
\000\000\186\000\000\000\186\000\186\000\186\000\000\000\186\000\
\186\000\186\000\186\000\186\000\191\000\191\000\191\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\191\000\000\000\191\000\000\000\191\000\191\000\191\000\
\191\000\191\000\000\000\000\000\000\000\191\000\000\000\000\000\
\191\000\191\000\000\000\191\000\191\000\191\000\000\000\191\000\
\000\000\191\000\000\000\000\000\000\000\191\000\191\000\191\000\
\191\000\176\000\191\000\000\000\000\000\000\000\000\000\000\000\
\191\000\000\000\000\000\000\000\000\000\000\000\191\000\191\000\
\191\000\191\000\191\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\191\000\000\000\000\000\000\000\000\000\191\000\
\191\000\191\000\191\000\191\000\191\000\000\000\000\000\191\000\
\000\000\000\000\191\000\000\000\000\000\191\000\191\000\000\000\
\000\000\191\000\000\000\191\000\191\000\191\000\000\000\191\000\
\191\000\191\000\191\000\191\000\159\000\159\000\159\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\159\000\000\000\159\000\000\000\159\000\159\000\159\000\
\159\000\159\000\000\000\000\000\000\000\159\000\152\000\000\000\
\159\000\159\000\000\000\159\000\159\000\159\000\000\000\159\000\
\000\000\159\000\000\000\000\000\000\000\159\000\159\000\159\000\
\159\000\000\000\159\000\000\000\000\000\000\000\000\000\000\000\
\159\000\000\000\000\000\000\000\000\000\000\000\159\000\159\000\
\159\000\159\000\159\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\159\000\000\000\000\000\000\000\000\000\159\000\
\159\000\159\000\159\000\159\000\159\000\000\000\000\000\159\000\
\000\000\000\000\159\000\000\000\000\000\159\000\159\000\000\000\
\000\000\159\000\000\000\159\000\159\000\000\000\000\000\159\000\
\159\000\159\000\159\000\159\000\152\000\152\000\152\000\176\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\152\000\000\000\152\000\000\000\152\000\152\000\152\000\
\152\000\152\000\000\000\000\000\000\000\152\000\000\000\000\000\
\152\000\152\000\000\000\152\000\152\000\152\000\000\000\152\000\
\000\000\152\000\000\000\000\000\000\000\152\000\152\000\152\000\
\152\000\000\000\152\000\000\000\000\000\000\000\000\000\000\000\
\152\000\000\000\000\000\000\000\000\000\000\000\000\000\152\000\
\000\000\000\000\152\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\152\000\000\000\000\000\000\000\000\000\152\000\
\152\000\152\000\152\000\152\000\152\000\000\000\000\000\152\000\
\176\000\000\000\152\000\000\000\000\000\152\000\152\000\000\000\
\000\000\152\000\000\000\152\000\176\000\176\000\176\000\152\000\
\152\000\152\000\152\000\152\000\000\000\000\000\000\000\000\000\
\000\000\176\000\000\000\176\000\000\000\176\000\176\000\176\000\
\000\000\176\000\000\000\000\000\000\000\176\000\000\000\000\000\
\176\000\176\000\000\000\176\000\176\000\176\000\000\000\000\000\
\000\000\176\000\000\000\000\000\000\000\176\000\176\000\176\000\
\176\000\000\000\176\000\000\000\000\000\000\000\000\000\000\000\
\176\000\000\000\000\000\000\000\000\000\000\000\176\000\000\000\
\176\000\176\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\176\000\000\000\152\000\000\000\000\000\176\000\
\176\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\152\000\152\000\152\000\000\000\176\000\052\000\000\000\
\000\000\176\000\052\000\176\000\176\000\176\000\152\000\052\000\
\152\000\000\000\152\000\152\000\152\000\052\000\152\000\052\000\
\000\000\000\000\152\000\000\000\000\000\152\000\152\000\000\000\
\152\000\152\000\152\000\052\000\052\000\052\000\152\000\000\000\
\000\000\052\000\152\000\152\000\152\000\152\000\000\000\152\000\
\000\000\000\000\000\000\000\000\000\000\152\000\000\000\000\000\
\000\000\000\000\052\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\111\000\000\000\000\000\000\000\152\000\
\000\000\000\000\000\000\000\000\152\000\152\000\000\000\000\000\
\000\000\000\000\176\000\176\000\176\000\000\000\000\000\000\000\
\000\000\000\000\152\000\000\000\000\000\000\000\152\000\176\000\
\152\000\176\000\152\000\176\000\176\000\176\000\000\000\176\000\
\000\000\000\000\000\000\176\000\000\000\000\000\176\000\176\000\
\000\000\176\000\176\000\176\000\000\000\000\000\000\000\176\000\
\000\000\000\000\000\000\176\000\176\000\176\000\176\000\000\000\
\176\000\000\000\000\000\000\000\000\000\000\000\176\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\176\000\000\000\000\000\000\000\000\000\000\000\
\176\000\000\000\000\000\000\000\000\000\176\000\176\000\000\000\
\000\000\000\000\000\000\176\000\176\000\176\000\000\000\000\000\
\000\000\000\000\000\000\176\000\000\000\000\000\000\000\176\000\
\176\000\176\000\176\000\176\000\176\000\176\000\176\000\000\000\
\176\000\000\000\000\000\000\000\176\000\000\000\000\000\176\000\
\176\000\000\000\176\000\176\000\176\000\000\000\000\000\000\000\
\176\000\000\000\000\000\000\000\176\000\176\000\176\000\176\000\
\000\000\176\000\000\000\000\000\000\000\000\000\000\000\176\000\
\000\000\000\000\000\000\000\000\000\000\176\000\000\000\176\000\
\176\000\154\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\176\000\000\000\000\000\000\000\000\000\176\000\176\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\152\000\
\152\000\152\000\000\000\000\000\176\000\000\000\000\000\000\000\
\176\000\160\001\176\000\176\000\152\000\000\000\152\000\000\000\
\152\000\152\000\152\000\162\001\152\000\000\000\000\000\000\000\
\152\000\000\000\000\000\152\000\152\000\000\000\152\000\152\000\
\152\000\000\000\000\000\000\000\152\000\000\000\000\000\000\000\
\152\000\152\000\152\000\152\000\000\000\152\000\000\000\000\000\
\000\000\163\001\000\000\152\000\000\000\000\000\000\000\000\000\
\142\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\164\001\165\001\166\001\167\001\152\000\000\000\000\000\
\000\000\000\000\152\000\152\000\000\000\000\000\111\000\111\000\
\111\000\000\000\168\001\169\001\170\001\171\001\000\000\000\000\
\152\000\000\000\000\000\111\000\152\000\111\000\152\000\111\000\
\111\000\111\000\000\000\111\000\000\000\000\000\000\000\111\000\
\000\000\000\000\111\000\111\000\000\000\111\000\111\000\111\000\
\000\000\000\000\004\000\111\000\000\000\000\000\000\000\111\000\
\111\000\111\000\111\000\000\000\111\000\000\000\000\000\000\000\
\000\000\000\000\111\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\048\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\111\000\000\000\000\000\000\000\
\000\000\111\000\111\000\000\000\000\000\176\000\176\000\176\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\111\000\
\000\000\000\000\176\000\111\000\176\000\111\000\176\000\176\000\
\176\000\000\000\176\000\000\000\000\000\000\000\176\000\000\000\
\000\000\176\000\176\000\000\000\176\000\176\000\176\000\000\000\
\000\000\000\000\176\000\000\000\000\000\000\000\176\000\176\000\
\176\000\176\000\000\000\176\000\066\000\000\000\000\000\000\000\
\000\000\176\000\000\000\000\000\048\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\176\000\000\000\000\000\000\000\000\000\
\176\000\176\000\000\000\000\000\154\000\154\000\154\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\176\000\000\000\
\000\000\154\000\176\000\154\000\176\000\154\000\154\000\154\000\
\000\000\154\000\000\000\000\000\000\000\154\000\000\000\000\000\
\154\000\154\000\000\000\154\000\154\000\154\000\000\000\000\000\
\000\000\154\000\000\000\000\000\000\000\154\000\154\000\154\000\
\154\000\000\000\154\000\000\000\000\000\000\000\000\000\000\000\
\154\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\152\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\066\000\154\000\000\000\000\000\000\000\000\000\154\000\
\154\000\000\000\000\000\142\000\142\000\142\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\154\000\000\000\000\000\
\142\000\154\000\142\000\154\000\142\000\142\000\142\000\000\000\
\142\000\000\000\000\000\000\000\142\000\000\000\000\000\142\000\
\142\000\000\000\142\000\142\000\142\000\000\000\000\000\000\000\
\142\000\000\000\000\000\000\000\142\000\142\000\142\000\142\000\
\000\000\142\000\000\000\000\000\000\000\000\000\000\000\142\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\004\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\142\000\004\000\004\000\000\000\000\000\142\000\142\000\
\000\000\000\000\000\000\000\000\000\000\000\000\004\000\004\000\
\004\000\048\000\048\000\000\000\142\000\000\000\000\000\000\000\
\142\000\000\000\142\000\048\000\000\000\000\000\004\000\000\000\
\004\000\000\000\000\000\000\000\000\000\000\000\000\000\048\000\
\000\000\048\000\000\000\000\000\000\000\000\000\000\000\004\000\
\000\000\000\000\000\000\048\000\004\000\004\000\000\000\048\000\
\000\000\048\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\004\000\004\000\004\000\
\000\000\004\000\004\000\000\000\004\000\048\000\048\000\160\001\
\004\000\000\000\066\000\066\000\000\000\000\000\000\000\000\000\
\161\001\162\001\048\000\048\000\066\000\000\000\048\000\048\000\
\048\000\000\000\048\000\048\000\048\000\048\000\000\000\000\000\
\066\000\048\000\066\000\000\000\000\000\000\000\000\000\000\000\
\048\000\000\000\048\000\000\000\066\000\000\000\000\000\163\001\
\066\000\000\000\066\000\000\000\000\000\000\000\000\000\000\000\
\048\000\000\000\048\000\000\000\000\000\000\000\000\000\164\001\
\165\001\166\001\167\001\000\000\000\000\000\000\066\000\066\000\
\000\000\000\000\000\000\000\000\000\000\000\000\048\000\048\000\
\168\001\169\001\170\001\171\001\000\000\000\000\000\000\066\000\
\066\000\066\000\000\000\066\000\066\000\000\000\066\000\048\000\
\048\000\048\000\066\000\048\000\048\000\176\001\048\000\152\000\
\152\000\000\000\048\000\000\000\000\000\000\000\000\000\066\000\
\066\000\152\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\066\000\000\000\000\000\000\000\152\000\000\000\152\000\
\000\000\000\000\000\000\000\000\000\000\066\000\000\000\066\000\
\000\000\000\000\000\000\000\000\000\000\152\000\000\000\152\000\
\000\000\000\000\000\000\000\000\000\000\066\000\000\000\066\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\152\000\152\000\000\000\000\000\000\000\
\000\000\000\000\000\000\066\000\066\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\152\000\152\000\152\000\000\000\
\152\000\152\000\000\000\152\000\066\000\066\000\066\000\152\000\
\066\000\066\000\000\000\066\000\220\000\221\000\222\000\066\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\223\000\000\000\224\000\000\000\225\000\226\000\227\000\
\000\000\228\000\000\000\000\000\000\000\160\001\000\000\000\000\
\229\000\230\000\000\000\231\000\232\000\233\000\161\001\162\001\
\120\001\000\000\121\001\000\000\000\000\234\000\235\000\236\000\
\237\000\000\000\238\000\160\001\000\000\000\000\000\000\160\001\
\239\000\000\000\000\000\000\000\161\001\162\001\000\000\000\000\
\161\001\162\001\000\000\000\000\000\000\163\001\000\000\000\000\
\000\000\000\000\240\000\000\000\000\000\000\000\000\000\241\000\
\242\000\000\000\000\000\000\000\000\000\164\001\165\001\166\001\
\167\001\000\000\000\000\163\001\000\000\243\000\000\000\163\001\
\000\000\244\000\000\000\245\000\000\000\000\000\168\001\169\001\
\170\001\171\001\160\001\164\001\165\001\166\001\167\001\164\001\
\165\001\166\001\167\001\161\001\162\001\000\000\000\000\000\000\
\000\000\000\000\000\000\122\001\168\001\169\001\170\001\171\001\
\168\001\169\001\170\001\171\001\160\001\006\000\007\000\000\000\
\000\000\000\000\000\000\000\000\000\000\161\001\162\001\008\000\
\000\000\177\001\163\001\000\000\000\000\178\001\000\000\000\000\
\000\000\000\000\000\000\009\000\000\000\010\000\000\000\000\000\
\000\000\000\000\164\001\165\001\166\001\167\001\000\000\000\000\
\000\000\000\000\000\000\011\000\163\001\012\000\000\000\000\000\
\000\000\069\001\000\000\168\001\169\001\170\001\171\001\000\000\
\000\000\000\000\000\000\000\000\164\001\165\001\166\001\167\001\
\070\001\013\000\014\000\000\000\000\000\000\000\000\000\000\000\
\183\001\000\000\000\000\000\000\000\000\168\001\169\001\170\001\
\171\001\000\000\015\000\016\000\017\000\000\000\018\000\019\000\
\071\001\020\000\000\000\072\001\000\000\021\000\160\001\000\000\
\000\000\000\000\224\001\000\000\000\000\160\001\223\001\161\001\
\162\001\000\000\073\001\074\001\075\001\076\001\161\001\162\001\
\077\001\000\000\000\000\078\001\000\000\000\000\000\000\079\001\
\000\000\187\001\000\000\000\000\000\000\000\000\160\001\000\000\
\080\001\081\001\082\001\083\001\084\001\160\001\163\001\161\001\
\162\001\000\000\000\000\000\000\203\001\163\001\161\001\162\001\
\000\000\000\000\000\000\000\000\000\000\000\000\164\001\165\001\
\166\001\167\001\000\000\000\000\000\000\164\001\165\001\166\001\
\167\001\000\000\000\000\000\000\000\000\000\000\163\001\168\001\
\169\001\170\001\171\001\000\000\000\000\163\001\168\001\169\001\
\170\001\171\001\000\000\000\000\000\000\000\000\164\001\165\001\
\166\001\167\001\000\000\000\000\000\000\164\001\165\001\166\001\
\167\001\000\000\000\000\000\000\000\000\000\000\000\000\168\001\
\169\001\170\001\171\001\000\000\000\000\000\000\168\001\169\001\
\170\001\171\001"

let yycheck = "\078\000\
\079\000\247\000\000\000\000\000\000\000\000\000\109\001\028\000\
\193\000\000\000\012\001\024\001\014\001\000\000\010\001\011\001\
\095\000\067\001\068\001\071\000\095\001\028\001\000\000\010\001\
\000\000\041\001\042\001\113\000\015\001\039\001\049\000\037\001\
\027\001\009\001\040\001\017\001\086\001\087\001\088\001\089\001\
\090\001\095\001\121\000\019\001\054\001\108\001\180\000\029\001\
\182\000\095\001\026\001\012\001\065\001\014\001\095\001\061\001\
\077\000\061\001\068\001\069\001\070\001\005\001\006\001\095\001\
\249\000\250\000\116\001\061\001\074\001\088\000\095\001\150\000\
\095\001\070\001\095\001\072\001\073\001\074\001\075\001\061\001\
\095\001\091\001\079\001\080\001\081\001\082\001\083\001\084\001\
\095\001\050\001\051\001\052\001\053\001\095\001\198\000\199\000\
\072\001\176\000\177\000\178\000\179\000\097\001\108\001\155\000\
\117\001\191\000\095\001\126\000\127\000\038\001\097\001\055\001\
\056\001\110\001\118\001\119\001\061\001\107\001\063\001\064\001\
\037\001\140\000\037\001\120\001\121\001\040\001\123\001\124\001\
\043\001\126\001\151\000\010\001\011\001\048\001\000\000\095\001\
\015\001\187\001\135\001\136\001\137\001\024\000\025\000\001\001\
\061\001\095\001\150\001\151\001\152\001\153\001\154\001\042\001\
\010\001\011\001\099\001\145\001\095\001\074\001\095\001\238\000\
\037\001\095\001\241\000\160\001\161\001\162\001\163\001\164\001\
\165\001\166\001\167\001\168\001\169\001\170\001\171\001\172\001\
\095\001\174\001\095\001\061\001\063\000\064\000\001\001\041\001\
\061\001\231\000\232\000\041\001\041\001\095\001\001\001\010\001\
\011\001\001\000\002\000\003\000\004\000\074\001\043\001\057\001\
\058\001\059\001\060\001\048\001\057\001\058\001\059\001\060\001\
\031\001\095\001\229\000\056\001\035\001\036\001\233\000\095\001\
\074\001\075\001\076\001\077\001\001\001\240\000\041\001\095\001\
\047\001\101\001\223\001\095\001\095\001\074\001\041\001\140\001\
\141\001\142\001\143\001\095\001\226\001\095\001\057\001\058\001\
\059\001\060\001\061\001\000\000\063\001\064\001\057\001\058\001\
\059\001\060\001\108\001\109\001\110\001\095\001\095\001\074\001\
\075\001\076\001\077\001\095\001\041\001\095\001\095\001\074\001\
\075\001\076\001\077\001\008\001\095\001\095\001\095\001\095\001\
\108\001\109\001\110\001\098\001\057\001\058\001\059\001\060\001\
\099\001\001\001\002\001\003\001\004\001\005\001\028\001\028\001\
\028\001\028\001\010\001\011\001\012\001\028\001\014\001\015\001\
\016\001\028\001\018\001\062\001\020\001\021\001\022\001\023\001\
\024\001\101\001\028\001\130\001\028\001\040\001\133\001\031\001\
\032\001\095\001\034\001\035\001\036\001\040\001\038\001\040\001\
\040\001\095\001\095\001\146\001\044\001\045\001\046\001\047\001\
\095\001\049\001\071\001\071\001\071\001\043\001\095\001\055\001\
\071\001\040\001\048\001\095\001\040\001\061\001\062\001\063\001\
\064\001\065\001\056\001\108\001\109\001\110\001\040\001\095\001\
\000\000\073\001\074\001\075\001\076\001\077\001\078\001\079\001\
\080\001\081\001\082\001\083\001\074\001\025\001\086\001\095\001\
\095\001\089\001\108\001\108\001\092\001\093\001\108\001\095\001\
\096\001\095\001\098\001\099\001\100\001\144\001\102\001\103\001\
\104\001\105\001\106\001\095\001\108\001\109\001\110\001\095\001\
\002\001\003\001\004\001\005\001\037\001\095\001\095\001\040\001\
\010\001\011\001\012\001\095\001\014\001\015\001\016\001\095\001\
\018\001\095\001\020\001\021\001\022\001\023\001\024\001\238\001\
\095\001\240\001\028\001\242\001\061\001\031\001\032\001\108\001\
\034\001\035\001\036\001\095\001\038\001\040\001\040\001\095\001\
\097\001\074\001\044\001\045\001\046\001\047\001\037\001\049\001\
\074\001\040\001\095\001\097\001\107\001\055\001\039\001\095\001\
\097\001\095\001\095\001\061\001\062\001\063\001\064\001\065\001\
\095\001\095\001\095\001\041\001\040\001\000\000\061\001\073\001\
\074\001\075\001\076\001\077\001\078\001\079\001\080\001\081\001\
\082\001\083\001\028\001\074\001\086\001\095\001\095\001\089\001\
\095\001\095\001\092\001\093\001\008\001\095\001\096\001\095\001\
\098\001\099\001\100\001\095\001\102\001\103\001\104\001\105\001\
\106\001\095\001\108\001\109\001\110\001\002\001\003\001\004\001\
\005\001\095\001\095\001\095\001\095\001\010\001\011\001\012\001\
\095\001\014\001\015\001\016\001\095\001\018\001\095\001\020\001\
\021\001\022\001\023\001\024\001\095\001\095\001\095\001\028\001\
\095\001\095\001\031\001\032\001\095\001\034\001\035\001\036\001\
\074\001\038\001\095\001\040\001\095\001\040\001\095\001\044\001\
\045\001\046\001\047\001\037\001\049\001\095\001\040\001\041\001\
\042\001\043\001\055\001\095\001\095\001\071\001\048\001\095\001\
\061\001\062\001\063\001\064\001\065\001\095\001\095\001\095\001\
\095\001\095\001\000\000\093\001\073\001\074\001\075\001\076\001\
\077\001\078\001\079\001\080\001\081\001\082\001\083\001\095\001\
\095\001\086\001\095\001\095\001\089\001\095\001\095\001\092\001\
\093\001\095\001\095\001\096\001\100\001\098\001\099\001\100\001\
\095\001\102\001\103\001\104\001\105\001\106\001\033\001\108\001\
\109\001\110\001\002\001\003\001\004\001\005\001\095\001\095\001\
\095\001\095\001\010\001\011\001\012\001\095\001\014\001\015\001\
\016\001\095\001\018\001\095\001\020\001\021\001\022\001\023\001\
\024\001\028\001\095\001\095\001\028\001\028\001\095\001\031\001\
\032\001\095\001\034\001\035\001\036\001\095\001\038\001\095\001\
\040\001\095\001\095\001\095\001\044\001\045\001\046\001\047\001\
\037\001\049\001\041\001\040\001\011\001\042\001\043\001\055\001\
\095\001\095\001\095\001\048\001\041\001\061\001\062\001\063\001\
\064\001\065\001\037\001\095\001\095\001\040\001\095\001\000\000\
\043\001\073\001\074\001\075\001\040\001\048\001\078\001\079\001\
\080\001\081\001\082\001\083\001\040\001\095\001\086\001\013\001\
\095\001\089\001\040\001\028\001\092\001\093\001\095\001\095\001\
\096\001\095\001\098\001\099\001\100\001\040\001\102\001\103\001\
\104\001\105\001\106\001\095\001\108\001\109\001\110\001\002\001\
\003\001\004\001\005\001\095\001\095\001\095\001\095\001\010\001\
\011\001\012\001\095\001\014\001\015\001\016\001\093\001\018\001\
\093\001\020\001\021\001\022\001\023\001\024\001\000\000\093\001\
\000\000\028\001\000\000\040\001\031\001\032\001\040\001\034\001\
\035\001\036\001\040\001\038\001\040\001\040\001\074\001\040\001\
\000\000\044\001\045\001\046\001\047\001\040\001\049\001\123\000\
\125\000\033\000\061\000\146\001\055\001\189\000\214\000\193\001\
\192\001\042\001\061\001\062\001\063\001\064\001\065\001\090\000\
\145\000\255\255\255\255\000\000\255\255\255\255\073\001\074\001\
\075\001\255\255\255\255\078\001\079\001\080\001\081\001\082\001\
\083\001\255\255\255\255\086\001\255\255\255\255\089\001\255\255\
\255\255\092\001\093\001\255\255\095\001\096\001\255\255\098\001\
\099\001\100\001\255\255\102\001\103\001\104\001\105\001\106\001\
\255\255\108\001\109\001\110\001\002\001\003\001\004\001\005\001\
\255\255\255\255\255\255\255\255\010\001\011\001\012\001\255\255\
\014\001\015\001\016\001\255\255\018\001\255\255\020\001\021\001\
\022\001\023\001\024\001\255\255\255\255\255\255\028\001\255\255\
\255\255\031\001\032\001\255\255\034\001\035\001\036\001\255\255\
\038\001\255\255\040\001\255\255\255\255\255\255\044\001\045\001\
\046\001\047\001\255\255\049\001\255\255\255\255\255\255\255\255\
\255\255\055\001\255\255\255\255\255\255\255\255\255\255\061\001\
\062\001\063\001\064\001\065\001\255\255\255\255\000\000\255\255\
\255\255\255\255\255\255\073\001\255\255\255\255\255\255\255\255\
\078\001\079\001\080\001\081\001\082\001\083\001\255\255\255\255\
\086\001\255\255\255\255\089\001\255\255\255\255\092\001\093\001\
\255\255\095\001\096\001\255\255\098\001\099\001\100\001\255\255\
\102\001\103\001\104\001\105\001\106\001\255\255\108\001\109\001\
\110\001\002\001\003\001\004\001\005\001\255\255\255\255\255\255\
\255\255\010\001\255\255\012\001\255\255\014\001\015\001\016\001\
\255\255\018\001\255\255\020\001\021\001\022\001\023\001\024\001\
\255\255\255\255\255\255\028\001\255\255\255\255\031\001\032\001\
\255\255\034\001\035\001\036\001\255\255\038\001\255\255\040\001\
\255\255\255\255\255\255\044\001\045\001\046\001\047\001\255\255\
\049\001\255\255\255\255\255\255\255\255\255\255\055\001\255\255\
\255\255\255\255\255\255\255\255\061\001\062\001\063\001\064\001\
\065\001\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
\073\001\255\255\255\255\255\255\255\255\078\001\079\001\080\001\
\081\001\082\001\083\001\255\255\255\255\086\001\255\255\255\255\
\089\001\255\255\255\255\092\001\093\001\255\255\095\001\096\001\
\255\255\098\001\099\001\100\001\255\255\102\001\103\001\104\001\
\105\001\106\001\255\255\108\001\109\001\110\001\003\001\004\001\
\005\001\006\001\007\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\016\001\255\255\018\001\255\255\020\001\
\021\001\022\001\023\001\024\001\255\255\255\255\255\255\028\001\
\255\255\030\001\031\001\032\001\255\255\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\042\001\043\001\044\001\
\045\001\046\001\047\001\048\001\049\001\255\255\255\255\255\255\
\255\255\054\001\055\001\056\001\255\255\255\255\255\255\255\255\
\061\001\062\001\063\001\064\001\065\001\066\001\067\001\068\001\
\069\001\070\001\000\000\255\255\073\001\074\001\255\255\255\255\
\255\255\078\001\079\001\080\001\081\001\082\001\083\001\084\001\
\085\001\086\001\087\001\088\001\089\001\090\001\091\001\092\001\
\093\001\094\001\095\001\096\001\255\255\098\001\099\001\100\001\
\255\255\102\001\103\001\104\001\105\001\106\001\255\255\255\255\
\109\001\003\001\004\001\005\001\006\001\007\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\016\001\255\255\
\018\001\255\255\020\001\021\001\022\001\023\001\024\001\255\255\
\255\255\255\255\028\001\255\255\030\001\031\001\032\001\255\255\
\034\001\035\001\036\001\037\001\038\001\039\001\040\001\041\001\
\042\001\043\001\044\001\045\001\046\001\047\001\048\001\049\001\
\255\255\255\255\255\255\255\255\054\001\055\001\056\001\255\255\
\255\255\255\255\255\255\061\001\062\001\063\001\064\001\065\001\
\066\001\067\001\068\001\069\001\070\001\000\000\255\255\073\001\
\074\001\255\255\255\255\255\255\078\001\079\001\080\001\081\001\
\082\001\083\001\084\001\085\001\086\001\087\001\088\001\089\001\
\090\001\091\001\092\001\093\001\094\001\255\255\096\001\255\255\
\098\001\099\001\100\001\255\255\102\001\103\001\104\001\105\001\
\106\001\255\255\255\255\109\001\003\001\004\001\005\001\006\001\
\007\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\030\001\
\031\001\032\001\255\255\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\042\001\043\001\044\001\045\001\046\001\
\047\001\048\001\049\001\255\255\255\255\255\255\255\255\054\001\
\055\001\056\001\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\066\001\067\001\068\001\069\001\070\001\
\000\000\255\255\073\001\074\001\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\084\001\085\001\086\001\
\087\001\088\001\089\001\090\001\091\001\092\001\093\001\094\001\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\255\255\255\255\109\001\255\255\
\255\255\255\255\255\255\255\255\255\255\003\001\004\001\005\001\
\006\001\007\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\016\001\255\255\018\001\255\255\020\001\021\001\
\022\001\023\001\024\001\255\255\255\255\255\255\028\001\255\255\
\030\001\031\001\032\001\255\255\034\001\035\001\036\001\037\001\
\038\001\255\255\040\001\255\255\255\255\255\255\044\001\045\001\
\046\001\047\001\255\255\049\001\255\255\255\255\255\255\255\255\
\255\255\055\001\255\255\001\001\255\255\255\255\255\255\061\001\
\062\001\000\000\255\255\065\001\066\001\067\001\255\255\255\255\
\255\255\071\001\255\255\073\001\074\001\255\255\255\255\255\255\
\078\001\079\001\080\001\081\001\082\001\083\001\084\001\085\001\
\086\001\087\001\088\001\089\001\090\001\255\255\092\001\093\001\
\094\001\255\255\096\001\041\001\098\001\255\255\100\001\255\255\
\102\001\103\001\104\001\105\001\106\001\255\255\108\001\255\255\
\003\001\004\001\005\001\057\001\058\001\059\001\060\001\255\255\
\255\255\255\255\255\255\255\255\255\255\016\001\255\255\018\001\
\255\255\020\001\021\001\022\001\023\001\024\001\076\001\077\001\
\255\255\028\001\255\255\255\255\031\001\032\001\033\001\034\001\
\035\001\036\001\255\255\038\001\255\255\040\001\255\255\255\255\
\255\255\044\001\045\001\046\001\047\001\255\255\049\001\255\255\
\255\255\255\255\255\255\255\255\055\001\255\255\255\255\255\255\
\255\255\000\000\061\001\062\001\063\001\064\001\065\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\073\001\255\255\
\255\255\255\255\255\255\078\001\079\001\080\001\081\001\082\001\
\083\001\255\255\255\255\086\001\255\255\255\255\089\001\255\255\
\255\255\092\001\093\001\255\255\095\001\096\001\255\255\098\001\
\099\001\255\255\255\255\102\001\103\001\104\001\105\001\106\001\
\255\255\255\255\255\255\003\001\004\001\005\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\016\001\255\255\018\001\255\255\020\001\021\001\022\001\023\001\
\024\001\255\255\255\255\255\255\028\001\255\255\255\255\031\001\
\032\001\255\255\034\001\035\001\036\001\255\255\038\001\255\255\
\040\001\255\255\255\255\255\255\044\001\045\001\046\001\047\001\
\255\255\049\001\255\255\255\255\255\255\255\255\255\255\055\001\
\255\255\000\000\255\255\255\255\255\255\061\001\062\001\063\001\
\064\001\065\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\073\001\255\255\255\255\255\255\255\255\078\001\079\001\
\080\001\081\001\082\001\083\001\255\255\255\255\086\001\255\255\
\255\255\089\001\255\255\255\255\092\001\093\001\255\255\095\001\
\096\001\255\255\098\001\099\001\255\255\255\255\102\001\103\001\
\104\001\105\001\106\001\255\255\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\095\001\096\001\255\255\098\001\099\001\255\255\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\000\000\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\000\000\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\255\255\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\100\001\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\000\000\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\255\255\255\255\255\255\255\255\255\255\061\001\062\001\
\063\001\064\001\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\255\255\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\099\001\255\255\255\255\102\001\
\103\001\104\001\105\001\106\001\003\001\004\001\005\001\000\000\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\023\001\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\038\001\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\255\255\255\255\255\255\255\255\255\255\255\255\062\001\
\255\255\255\255\065\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\080\001\081\001\082\001\083\001\255\255\255\255\086\001\
\000\000\255\255\089\001\255\255\255\255\092\001\093\001\255\255\
\255\255\096\001\255\255\098\001\003\001\004\001\005\001\102\001\
\103\001\104\001\105\001\106\001\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\255\255\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\255\255\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\255\255\255\255\255\255\255\255\255\255\061\001\255\255\
\063\001\064\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\073\001\255\255\000\000\255\255\255\255\078\001\
\079\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\003\001\004\001\005\001\255\255\092\001\039\001\255\255\
\255\255\096\001\043\001\098\001\099\001\100\001\016\001\048\001\
\018\001\255\255\020\001\021\001\022\001\054\001\024\001\056\001\
\255\255\255\255\028\001\255\255\255\255\031\001\032\001\255\255\
\034\001\035\001\036\001\068\001\069\001\070\001\040\001\255\255\
\255\255\074\001\044\001\045\001\046\001\047\001\255\255\049\001\
\255\255\255\255\255\255\255\255\255\255\055\001\255\255\255\255\
\255\255\255\255\091\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\000\000\255\255\255\255\255\255\073\001\
\255\255\255\255\255\255\255\255\078\001\079\001\255\255\255\255\
\255\255\255\255\003\001\004\001\005\001\255\255\255\255\255\255\
\255\255\255\255\092\001\255\255\255\255\255\255\096\001\016\001\
\098\001\018\001\100\001\020\001\021\001\022\001\255\255\024\001\
\255\255\255\255\255\255\028\001\255\255\255\255\031\001\032\001\
\255\255\034\001\035\001\036\001\255\255\255\255\255\255\040\001\
\255\255\255\255\255\255\044\001\045\001\046\001\047\001\255\255\
\049\001\255\255\255\255\255\255\255\255\255\255\055\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
\073\001\255\255\255\255\255\255\255\255\078\001\079\001\255\255\
\255\255\255\255\255\255\003\001\004\001\005\001\255\255\255\255\
\255\255\255\255\255\255\092\001\255\255\255\255\255\255\096\001\
\016\001\098\001\018\001\100\001\020\001\021\001\022\001\255\255\
\024\001\255\255\255\255\255\255\028\001\255\255\255\255\031\001\
\032\001\255\255\034\001\035\001\036\001\255\255\255\255\255\255\
\040\001\255\255\255\255\255\255\044\001\045\001\046\001\047\001\
\255\255\049\001\255\255\255\255\255\255\255\255\255\255\055\001\
\255\255\255\255\255\255\255\255\255\255\061\001\255\255\063\001\
\064\001\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\073\001\255\255\255\255\255\255\255\255\078\001\079\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\003\001\
\004\001\005\001\255\255\255\255\092\001\255\255\255\255\255\255\
\096\001\001\001\098\001\099\001\016\001\255\255\018\001\255\255\
\020\001\021\001\022\001\011\001\024\001\255\255\255\255\255\255\
\028\001\255\255\255\255\031\001\032\001\255\255\034\001\035\001\
\036\001\255\255\255\255\255\255\040\001\255\255\255\255\255\255\
\044\001\045\001\046\001\047\001\255\255\049\001\255\255\255\255\
\255\255\041\001\255\255\055\001\255\255\255\255\255\255\255\255\
\000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\057\001\058\001\059\001\060\001\073\001\255\255\255\255\
\255\255\255\255\078\001\079\001\255\255\255\255\003\001\004\001\
\005\001\255\255\074\001\075\001\076\001\077\001\255\255\255\255\
\092\001\255\255\255\255\016\001\096\001\018\001\098\001\020\001\
\021\001\022\001\255\255\024\001\255\255\255\255\255\255\028\001\
\255\255\255\255\031\001\032\001\255\255\034\001\035\001\036\001\
\255\255\255\255\000\000\040\001\255\255\255\255\255\255\044\001\
\045\001\046\001\047\001\255\255\049\001\255\255\255\255\255\255\
\255\255\255\255\055\001\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\073\001\255\255\255\255\255\255\
\255\255\078\001\079\001\255\255\255\255\003\001\004\001\005\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\092\001\
\255\255\255\255\016\001\096\001\018\001\098\001\020\001\021\001\
\022\001\255\255\024\001\255\255\255\255\255\255\028\001\255\255\
\255\255\031\001\032\001\255\255\034\001\035\001\036\001\255\255\
\255\255\255\255\040\001\255\255\255\255\255\255\044\001\045\001\
\046\001\047\001\255\255\049\001\000\000\255\255\255\255\255\255\
\255\255\055\001\255\255\255\255\000\000\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\073\001\255\255\255\255\255\255\255\255\
\078\001\079\001\255\255\255\255\003\001\004\001\005\001\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\092\001\255\255\
\255\255\016\001\096\001\018\001\098\001\020\001\021\001\022\001\
\255\255\024\001\255\255\255\255\255\255\028\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\255\255\255\255\
\255\255\040\001\255\255\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\255\255\255\255\255\255\255\255\255\255\
\055\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\000\000\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\255\255\255\255\003\001\004\001\005\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\092\001\255\255\255\255\
\016\001\096\001\018\001\098\001\020\001\021\001\022\001\255\255\
\024\001\255\255\255\255\255\255\028\001\255\255\255\255\031\001\
\032\001\255\255\034\001\035\001\036\001\255\255\255\255\255\255\
\040\001\255\255\255\255\255\255\044\001\045\001\046\001\047\001\
\255\255\049\001\255\255\255\255\255\255\255\255\255\255\055\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\006\001\007\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\073\001\016\001\017\001\255\255\255\255\078\001\079\001\
\255\255\255\255\255\255\255\255\255\255\255\255\028\001\029\001\
\030\001\006\001\007\001\255\255\092\001\255\255\255\255\255\255\
\096\001\255\255\098\001\016\001\255\255\255\255\044\001\255\255\
\046\001\255\255\255\255\255\255\255\255\255\255\255\255\028\001\
\255\255\030\001\255\255\255\255\255\255\255\255\255\255\061\001\
\255\255\255\255\255\255\040\001\066\001\067\001\255\255\044\001\
\255\255\046\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\083\001\084\001\085\001\
\255\255\087\001\088\001\255\255\090\001\066\001\067\001\001\001\
\094\001\255\255\006\001\007\001\255\255\255\255\255\255\255\255\
\010\001\011\001\006\001\007\001\016\001\255\255\083\001\084\001\
\085\001\255\255\087\001\088\001\016\001\090\001\255\255\255\255\
\028\001\094\001\030\001\255\255\255\255\255\255\255\255\255\255\
\028\001\255\255\030\001\255\255\040\001\255\255\255\255\041\001\
\044\001\255\255\046\001\255\255\255\255\255\255\255\255\255\255\
\044\001\255\255\046\001\255\255\255\255\255\255\255\255\057\001\
\058\001\059\001\060\001\255\255\255\255\255\255\066\001\067\001\
\255\255\255\255\255\255\255\255\255\255\255\255\066\001\067\001\
\074\001\075\001\076\001\077\001\255\255\255\255\255\255\083\001\
\084\001\085\001\255\255\087\001\088\001\255\255\090\001\083\001\
\084\001\085\001\094\001\087\001\088\001\095\001\090\001\006\001\
\007\001\255\255\094\001\255\255\255\255\255\255\255\255\006\001\
\007\001\016\001\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\255\255\255\255\028\001\255\255\030\001\
\255\255\255\255\255\255\255\255\255\255\028\001\255\255\030\001\
\255\255\255\255\255\255\255\255\255\255\044\001\255\255\046\001\
\255\255\255\255\255\255\255\255\255\255\044\001\255\255\046\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\255\255\255\255\066\001\067\001\255\255\255\255\255\255\
\255\255\255\255\255\255\066\001\067\001\255\255\255\255\255\255\
\255\255\255\255\255\255\255\255\083\001\084\001\085\001\255\255\
\087\001\088\001\255\255\090\001\083\001\084\001\085\001\094\001\
\087\001\088\001\255\255\090\001\003\001\004\001\005\001\094\001\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
\255\255\016\001\255\255\018\001\255\255\020\001\021\001\022\001\
\255\255\024\001\255\255\255\255\255\255\001\001\255\255\255\255\
\031\001\032\001\255\255\034\001\035\001\036\001\010\001\011\001\
\012\001\255\255\014\001\255\255\255\255\044\001\045\001\046\001\
\047\001\255\255\049\001\001\001\255\255\255\255\255\255\001\001\
\055\001\255\255\255\255\255\255\010\001\011\001\255\255\255\255\
\010\001\011\001\255\255\255\255\255\255\041\001\255\255\255\255\
\255\255\255\255\073\001\255\255\255\255\255\255\255\255\078\001\
\079\001\255\255\255\255\255\255\255\255\057\001\058\001\059\001\
\060\001\255\255\255\255\041\001\255\255\092\001\255\255\041\001\
\255\255\096\001\255\255\098\001\255\255\255\255\074\001\075\001\
\076\001\077\001\001\001\057\001\058\001\059\001\060\001\057\001\
\058\001\059\001\060\001\010\001\011\001\255\255\255\255\255\255\
\255\255\255\255\255\255\095\001\074\001\075\001\076\001\077\001\
\074\001\075\001\076\001\077\001\001\001\006\001\007\001\255\255\
\255\255\255\255\255\255\255\255\255\255\010\001\011\001\016\001\
\255\255\095\001\041\001\255\255\255\255\095\001\255\255\255\255\
\255\255\255\255\255\255\028\001\255\255\030\001\255\255\255\255\
\255\255\255\255\057\001\058\001\059\001\060\001\255\255\255\255\
\255\255\255\255\255\255\044\001\041\001\046\001\255\255\255\255\
\255\255\023\001\255\255\074\001\075\001\076\001\077\001\255\255\
\255\255\255\255\255\255\255\255\057\001\058\001\059\001\060\001\
\038\001\066\001\067\001\255\255\255\255\255\255\255\255\255\255\
\095\001\255\255\255\255\255\255\255\255\074\001\075\001\076\001\
\077\001\255\255\083\001\084\001\085\001\255\255\087\001\088\001\
\062\001\090\001\255\255\065\001\255\255\094\001\001\001\255\255\
\255\255\255\255\095\001\255\255\255\255\001\001\002\001\010\001\
\011\001\255\255\080\001\081\001\082\001\083\001\010\001\011\001\
\086\001\255\255\255\255\089\001\255\255\255\255\255\255\093\001\
\255\255\028\001\255\255\255\255\255\255\255\255\001\001\255\255\
\102\001\103\001\104\001\105\001\106\001\001\001\041\001\010\001\
\011\001\255\255\255\255\255\255\015\001\041\001\010\001\011\001\
\255\255\255\255\255\255\255\255\255\255\255\255\057\001\058\001\
\059\001\060\001\255\255\255\255\255\255\057\001\058\001\059\001\
\060\001\255\255\255\255\255\255\255\255\255\255\041\001\074\001\
\075\001\076\001\077\001\255\255\255\255\041\001\074\001\075\001\
\076\001\077\001\255\255\255\255\255\255\255\255\057\001\058\001\
\059\001\060\001\255\255\255\255\255\255\057\001\058\001\059\001\
\060\001\255\255\255\255\255\255\255\255\255\255\255\255\074\001\
\075\001\076\001\077\001\255\255\255\255\255\255\074\001\075\001\
\076\001\077\001"

let yynames_const = "\
  QUESTION\000\
  COLON\000\
  ACTION_IF\000\
  ADD_KIT\000\
  ADD_MUSIC\000\
  ADD_STATE_TRIGGER\000\
  ADD_TRANS_TRIGGER\000\
  ALLOW_MISSING\000\
  ALWAYS\000\
  AND\000\
  OR\000\
  NOT\000\
  ELSE\000\
  LPAREN\000\
  RPAREN\000\
  APPEND\000\
  APPENDI\000\
  APPEND_COL\000\
  ASK_EVERY_COMPONENT\000\
  AT_EXIT\000\
  AT_INTERACTIVE_EXIT\000\
  AT_INTERACTIVE_UNINSTALL\000\
  SET_2DA_ENTRY\000\
  AT_UNINSTALL\000\
  AUTHOR\000\
  AUTO_TRA\000\
  BACKUP\000\
  BEGIN\000\
  CHAIN2\000\
  CHAIN3\000\
  COMPILE\000\
  COPY\000\
  GLOB\000\
  COPY_EXISTING\000\
  COPY_EXISTING_REGEXP\000\
  COPY_RANDOM\000\
  COPY_TRANS\000\
  DELETE_BYTES\000\
  DO\000\
  END\000\
  EQUALS\000\
  EQUALSEQUALS\000\
  EXIT\000\
  EXTEND_BOTTOM\000\
  EXTEND_BOTTOM_REGEXP\000\
  EXTEND_TOP\000\
  EXTEND_TOP_REGEXP\000\
  EXTERN\000\
  FAIL\000\
  FILE_CONTAINS\000\
  FILE_EXISTS\000\
  FILE_EXISTS_IN_GAME\000\
  FILE_SIZE\000\
  FLAGS\000\
  FORBID_FILE\000\
  GOTO\000\
  GT\000\
  GTE\000\
  LT\000\
  LTE\000\
  IF\000\
  SET\000\
  IF_EVAL\000\
  IF_SIZE_IS\000\
  INSERT_BYTES\000\
  INTERJECT\000\
  INTERJECT_COPY_TRANS\000\
  JOURNAL\000\
  SOLVED_JOURNAL\000\
  UNSOLVED_JOURNAL\000\
  LANGUAGE\000\
  MENU_STYLE\000\
  MKDIR\000\
  PLUS\000\
  MINUS\000\
  TIMES\000\
  DIVIDE\000\
  PRINT\000\
  RANDOM_SEED\000\
  READ_BYTE\000\
  READ_LONG\000\
  READ_SHORT\000\
  REPLACE\000\
  REPLACE_ACTION_TEXT\000\
  REPLACE_ACTION_TEXT_PROCESS\000\
  REPLACE_BCS_BLOCK\000\
  REPLACE_SAY\000\
  REPLACE_STATE_TRIGGER\000\
  REPLACE_TEXTUALLY\000\
  REPLACE_TRIGGER_TEXT\000\
  REPLY\000\
  REQUIRE_FILE\000\
  SAY\000\
  SET_WEIGHT\000\
  STRING_SET\000\
  THEN\000\
  UNINSTALL\000\
  UNLESS\000\
  USING\000\
  WEIGHT\000\
  WRITE_ASCII\000\
  WRITE_BYTE\000\
  WRITE_LONG\000\
  WRITE_SHORT\000\
  WHILE\000\
  EOF\000\
  "

let yynames_block = "\
  STRING\000\
  SOUND\000\
  STRING_REF\000\
  TRANS_REF\000\
  FORCED_STRING_REF\000\
  "

let yyact = [|
  (fun _ -> failwith "parser")
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : Dc.action list) in
    Obj.repr((
# 210 "src/dparser.mly"
                         let answer = _1 @ !extra_actions in 
                         extra_actions := [] ;
                         answer ) : Dc.action list))
; (fun parser_env ->
    Obj.repr((
# 215 "src/dparser.mly"
                          [] ) : Dc.action list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : Dc.action) in
    let _2 = (peek_val parser_env 0 : Dc.action list) in
    Obj.repr((
# 216 "src/dparser.mly"
                          _1 :: _2 ) : Dc.action list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 220 "src/dparser.mly"
                             current_unit := Some(String.uppercase _2); (String.uppercase _2,0) ) : (string * int)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 221 "src/dparser.mly"
                              current_unit := Some(String.uppercase _2); (String.uppercase _2,my_int_of_string _3) ) : (string * int)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 225 "src/dparser.mly"
        current_unit := Some(String.uppercase _2) ; 
        (String.uppercase _2) ) : (string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string list) in
    let _4 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 230 "src/dparser.mly"
      current_unit := Some(String.uppercase _2) ; 
    (true,String.uppercase _2,_3,_4) ) : (bool * string * (string list) * int)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string list) in
    let _4 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 233 "src/dparser.mly"
      current_unit := Some(String.uppercase _2) ;
    (false,String.uppercase _2,_3,_4) ) : (bool * string * (string list) * int)))
; (fun parser_env ->
    Obj.repr((
# 237 "src/dparser.mly"
                                0 ) : int))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 238 "src/dparser.mly"
                                _1 ) : int))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 242 "src/dparser.mly"
      current_unit := Some(String.uppercase _2); (String.uppercase _2) ) : (string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : string) in
    let _5 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 246 "src/dparser.mly"
    current_unit := Some(String.uppercase _2); 
    (String.uppercase _2,_3 :: _5,_4) ) : (string * (string list) * string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : string) in
    let _5 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 251 "src/dparser.mly"
    current_unit := Some(String.uppercase _2); 
    (String.uppercase _2,_3 :: _5,_4) ) : (string * (string list) * string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : string) in
    let _5 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 255 "src/dparser.mly"
    current_unit := Some(String.uppercase _2); 
    (String.uppercase _2,_3 :: _5,_4) 
  ) : (string * (string list) * string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : 'optional_weighted_condition) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 261 "src/dparser.mly"
    current_unit := Some(String.uppercase _3); (_2,String.uppercase _3,_4) ) : 'chain3_prologue))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 265 "src/dparser.mly"
    current_unit := Some(String.uppercase _2); (String.uppercase _2,_3,_4) ) : (string * string * string)))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 269 "src/dparser.mly"
    current_unit := Some(String.uppercase _2); (String.uppercase _2,_3,_4) ) : (string * string * string)))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : (string * int)) in
    let _2 = (peek_val parser_env 0 : Dlg.state list) in
    Obj.repr((
# 273 "src/dparser.mly"
                            let name,flags = _1 in Dc.Create(
  { Dlg.name = name ;
    Dlg.state = Array.of_list _2 ;
    Dlg.dlg_flags = flags ; })) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : (string)) in
    let _2 = (peek_val parser_env 1 : Dlg.state list) in
    Obj.repr((
# 278 "src/dparser.mly"
    current_unit := None ; Dc.Append( _1 , _2 ) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : (bool * string * (string list) * int)) in
    let _2 = (peek_val parser_env 1 : Dlg.transition list) in
    Obj.repr((
# 280 "src/dparser.mly"
    let top,ext_unit,ext_label,number = _1 in
    current_unit := None ; 
    if top then 
      Dc.Extend_Top( String.uppercase ext_unit, ext_label, number, _2 ) 
    else
      Dc.Extend_Bottom( String.uppercase ext_unit, ext_label, number, _2 ) 
  ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : (string)) in
    let _2 = (peek_val parser_env 1 : Dlg.state list) in
    Obj.repr((
# 288 "src/dparser.mly"
    current_unit := None ; 
    Dc.Replace(_1, _2) ) : Dc.action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 291 "src/dparser.mly"
    Dc.Replace_Say(_2,_3,_4) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : (string * (string list) * string)) in
    Obj.repr((
# 293 "src/dparser.mly"
    let f,s,t = _1 in 
    let verified_t = verify_trigger_list t in 
    current_unit := None ; 
    Dc.Replace_State_Trigger(f,s,verified_t) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : (string * (string list) * string)) in
    Obj.repr((
# 298 "src/dparser.mly"
    let f,s,t = _1 in 
    let verified_t = verify_trigger_list t in 
    current_unit := None ; 
    Dc.Add_State_Trigger(f,s,verified_t) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : (string * (string list) * string)) in
    Obj.repr((
# 303 "src/dparser.mly"
    let f,s,t = _1 in 
    let verified_t = verify_trigger_list t in 
    current_unit := None ; 
    Dc.Add_Trans_Trigger(f,s,verified_t) ) : Dc.action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 308 "src/dparser.mly"
    Dc.Set_Weight(_2,_3,_4) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 3 : 'chain3_prologue) in
    let _2 = (peek_val parser_env 2 : ((string option) * Dlg.tlk_string * (string option)) list) in
    let _3 = (peek_val parser_env 1 : 'compound_chain3_list) in
    let _4 = (peek_val parser_env 0 : 'chain3_epilogue) in
    Obj.repr((
# 310 "src/dparser.mly"
    let (entry_weight,entry_cond),file,state = _1 in 
    let first_part = List.map (fun (cond,says,action) -> 
      {
        Dc.c3du_speaker = file ;
        Dc.c3du_condition = cond ;
        Dc.c3du_action = action ;
        Dc.c3du_say = says 
      } ) _2 in 
    Dc.Chain3
    {
      Dc.c3_entry_condition = entry_cond;
      Dc.c3_entry_weight = entry_weight;
      Dc.c3_entry_file = file ;
      Dc.c3_entry_label = state ;
      Dc.c3_dialogue = first_part @ _3 ;
      Dc.c3_variable = None ; 
      Dc.c3_exit_trans = _4 ; 
    } 
  ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 4 : (string * string * string)) in
    let _2 = (peek_val parser_env 3 : 'compound_chain3_list) in
    let _4 = (peek_val parser_env 1 : string) in
    let _5 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 330 "src/dparser.mly"
    let file,label,var = _1 in 
    Dc.Chain3
    {
      Dc.c3_entry_condition = None;
      Dc.c3_entry_weight = Dlg.Not_Specified;
      Dc.c3_entry_file = file ;
      Dc.c3_entry_label = label ;
      Dc.c3_dialogue = _2 ;
      Dc.c3_variable = Some(var) ; 
      Dc.c3_exit_trans = [| Dlg.make_trans_of_next (Dlg.Symbolic(_4,_5)) |] ; 
    } 
  ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : (string * string * string)) in
    let _2 = (peek_val parser_env 1 : 'compound_chain3_list) in
    Obj.repr((
# 343 "src/dparser.mly"
    let file,label,var = _1 in 
    Dc.Chain3
    {
      Dc.c3_entry_condition = None;
      Dc.c3_entry_weight = Dlg.Not_Specified;
      Dc.c3_entry_file = file ;
      Dc.c3_entry_label = label ;
      Dc.c3_dialogue = _2 ;
      Dc.c3_variable = Some(var) ; 
      Dc.c3_exit_trans = [| Dlg.make_trans_of_next (Dlg.Copy(file,label)) |] 
    } 
  ) : Dc.action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 356 "src/dparser.mly"
    Dc.Replace_Trigger_Text(String.uppercase _2,_3,_4) ) : Dc.action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 358 "src/dparser.mly"
    Dc.Replace_Action_Text(String.uppercase _2,_3,_4) ) : Dc.action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 360 "src/dparser.mly"
    Dc.Replace_Action_Text(String.uppercase _2,_3,verify_action_list _4) ) : Dc.action))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : string option) in
    let _2 = (peek_val parser_env 1 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 0 : 'optional_action) in
    Obj.repr((
# 363 "src/dparser.mly"
                                                           [(_1,_2,_3)] ) : ((string option) * Dlg.tlk_string * (string option)) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 4 : string option) in
    let _2 = (peek_val parser_env 3 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 2 : 'optional_action) in
    let _5 = (peek_val parser_env 0 : ((string option) * Dlg.tlk_string * (string option)) list) in
    Obj.repr((
# 365 "src/dparser.mly"
    (_1,_2,_3) :: (List.map (fun (a,b,c) -> 
    match _1,a with
    | _,None -> (_1,b,c)
    | None,_ -> (a,b,c)
    | Some(first),Some(second) -> (Some(first ^ "\n" ^ second),b,c)
    ) _5) ) : ((string option) * Dlg.tlk_string * (string option)) list))
; (fun parser_env ->
    Obj.repr((
# 373 "src/dparser.mly"
                          None ) : 'optional_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 374 "src/dparser.mly"
                          let verified_action = verify_action_list _2 in 
                          Some(verified_action) ) : 'optional_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 379 "src/dparser.mly"
      let next = Dlg.Symbolic(String.uppercase _2,_3) in
      [| Dlg.make_trans_of_next next |] ) : 'chain3_epilogue))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 382 "src/dparser.mly"
      let next = Dlg.Symbolic(String.uppercase _2,_3) in
      [| Dlg.make_trans_of_next next |] ) : 'chain3_epilogue))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 385 "src/dparser.mly"
      let next = Dlg.Copy(String.uppercase _2,_3) in 
      [| Dlg.make_trans_of_next next |] ) : 'chain3_epilogue))
; (fun parser_env ->
    Obj.repr((
# 388 "src/dparser.mly"
      let next = Dlg.Exit in 
      [| Dlg.make_trans_of_next next |] ) : 'chain3_epilogue))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.transition list) in
    Obj.repr((
# 390 "src/dparser.mly"
                                       Array.of_list _2 ) : 'chain3_epilogue))
; (fun parser_env ->
    Obj.repr((
# 393 "src/dparser.mly"
                                   ((Dlg.Not_Specified),(None)) ) : 'optional_weighted_condition))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : Dlg.weight) in
    let _3 = (peek_val parser_env 1 : string) in
    Obj.repr((
# 394 "src/dparser.mly"
                                 (_2,(Some(verify_trigger_list _3))) ) : 'optional_weighted_condition))
; (fun parser_env ->
    Obj.repr((
# 397 "src/dparser.mly"
                          None ) : string option))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    Obj.repr((
# 398 "src/dparser.mly"
                          Some(verify_trigger_list _2) ) : string option))
; (fun parser_env ->
    Obj.repr((
# 401 "src/dparser.mly"
                                          [] ) : 'compound_chain3_list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : ((string option) * Dlg.tlk_string * (string option)) list) in
    let _4 = (peek_val parser_env 0 : 'compound_chain3_list) in
    Obj.repr((
# 403 "src/dparser.mly"
    let this_speaker = String.uppercase _2 in 
    let first_part = List.map (fun (cond,says,action) -> 
      { Dc.c3du_speaker = this_speaker ;
        Dc.c3du_condition = cond ;
        Dc.c3du_action = action ;
        Dc.c3du_say = says }) _3 in 
    first_part @ _4 
  ) : 'compound_chain3_list))
; (fun parser_env ->
    Obj.repr((
# 415 "src/dparser.mly"
                          [] ) : Dlg.state list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : Dlg.state list) in
    let _2 = (peek_val parser_env 0 : Dlg.state list) in
    Obj.repr((
# 416 "src/dparser.mly"
                          _1 @ _2 ) : Dlg.state list))
; (fun parser_env ->
    Obj.repr((
# 419 "src/dparser.mly"
                          Dlg.Not_Specified ) : Dlg.weight))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 420 "src/dparser.mly"
                          Dlg.Offset(_2) ) : Dlg.weight))
; (fun parser_env ->
    Obj.repr((
# 423 "src/dparser.mly"
                 () ) : 'optional_then))
; (fun parser_env ->
    Obj.repr((
# 424 "src/dparser.mly"
         () ) : 'optional_then))
; (fun parser_env ->
    Obj.repr((
# 426 "src/dparser.mly"
                  () ) : 'optional_begin))
; (fun parser_env ->
    Obj.repr((
# 427 "src/dparser.mly"
          () ) : 'optional_begin))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 431 "src/dparser.mly"
         if _1 = "" then "" else verify_trigger_list _1 ) : 'state_trigger))
; (fun parser_env ->
    let _2 = (peek_val parser_env 8 : Dlg.weight) in
    let _3 = (peek_val parser_env 7 : 'state_trigger) in
    let _4 = (peek_val parser_env 6 : 'optional_then) in
    let _5 = (peek_val parser_env 5 : 'optional_begin) in
    let _6 = (peek_val parser_env 4 : string) in
    let _8 = (peek_val parser_env 2 : Dlg.tlk_string list) in
    let _9 = (peek_val parser_env 1 : Dlg.transition list) in
    Obj.repr((
# 438 "src/dparser.mly"
     
      let state_trigger = _3 in 
      if List.length _8 = 1 then 
         [{ Dlg.resp_str = List.hd _8 ;
           Dlg.trans = Array.of_list _9 ; 
           Dlg.state_trigger = state_trigger ;
           Dlg.state_trigger_weight = _2 ; 
           Dlg.symbolic_label = _6 ; }]
      else begin
        let rec process_say_list sl in_label in_trig in_weight = match sl with
          s1 :: s2 :: tl -> 
            let new_label = Dc.chain_label () in
            let dest_file = get_current_unit () in 
            let new_state = Dlg.make_state s1 in_label dest_file new_label in
            new_state.Dlg.state_trigger_weight <- in_weight ; 
            new_state.Dlg.state_trigger <- in_trig ; 
            new_state :: (process_say_list (s2::tl) new_label "" Dlg.Not_Specified)
        | s1 :: [] -> [
          let state = Dlg.make_state_trans s1 in_label (Array.of_list _9)
          in state.Dlg.state_trigger_weight <- in_weight ;
          state ] 
        | [] -> [] 
        in 
        process_say_list _8 _6 _3 _2
      end
    ) : Dlg.state list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : string option * string) in
    let _2 = (peek_val parser_env 1 : Dlg.state list) in
    Obj.repr((
# 466 "src/dparser.mly"
   
    let old_unit, append_name = _1 in 
    extra_actions := (Dc.Append( append_name, _2 )) :: !extra_actions ;
    current_unit := old_unit ;
    [] 
  ) : Dlg.state list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 5 : string) in
    let _3 = (peek_val parser_env 4 : string) in
    let _4 = (peek_val parser_env 3 : (string * Dlg.tlk_string) list) in
    let _6 = (peek_val parser_env 1 : string) in
    let _7 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 474 "src/dparser.mly"
    let rec convert cl extern_guy = match cl with
      [] -> []
    | (code,s) :: tl -> 
        let file = if extern_guy then _2 else get_current_unit () in
        let extern_guy = if code = "==" then not extern_guy else extern_guy in
        (file,s) :: convert tl extern_guy
    in 
    extra_actions := (Dc.Chain(
      { Dc.entry_file = String.uppercase _2 ;
        Dc.entry_label = _3 ;
        Dc.dialogue = convert _4 true;
        Dc.exit_file = String.uppercase _6 ;
        Dc.exit_label = _7 ;
      } )) :: !extra_actions ;
    []  
  ) : Dlg.state list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 493 "src/dparser.mly"
   let what = String.uppercase _2 in  
  let old = !current_unit in current_unit := Some(what); (old, what) ) : string option * string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 497 "src/dparser.mly"
                                  [("",_1)] ) : (string * Dlg.tlk_string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 0 : (string * Dlg.tlk_string) list) in
    Obj.repr((
# 498 "src/dparser.mly"
                                  ("",_1) :: _3 ) : (string * Dlg.tlk_string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 0 : (string * Dlg.tlk_string) list) in
    Obj.repr((
# 499 "src/dparser.mly"
                                  ("==",_1) :: _3 ) : (string * Dlg.tlk_string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 502 "src/dparser.mly"
                                  [_1] ) : Dlg.tlk_string list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string list) in
    Obj.repr((
# 503 "src/dparser.mly"
                                  _1 :: _3 ) : Dlg.tlk_string list))
; (fun parser_env ->
    Obj.repr((
# 506 "src/dparser.mly"
                               [] ) : Dlg.transition list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : Dlg.transition) in
    let _2 = (peek_val parser_env 0 : Dlg.transition list) in
    Obj.repr((
# 507 "src/dparser.mly"
                               _1 :: _2 ) : Dlg.transition list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 511 "src/dparser.mly"
                   Trans_Reply(Some(_2)) ) : trans_opt_type))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 512 "src/dparser.mly"
                   let verified_action = verify_action_list _2 in 
                    Trans_Do(Some(verified_action)) ) : trans_opt_type))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 514 "src/dparser.mly"
                   Trans_Journal(Some(Dlg.Normal_Journal,_2)) ) : trans_opt_type))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 515 "src/dparser.mly"
                          Trans_Journal(Some(Dlg.Solved_Journal,_2)) ) : trans_opt_type))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 516 "src/dparser.mly"
                            Trans_Journal(Some(Dlg.Unsolved_Journal,_2)) ) : trans_opt_type))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 517 "src/dparser.mly"
                   Trans_Flags(my_int_of_string _2)) : trans_opt_type))
; (fun parser_env ->
    Obj.repr((
# 520 "src/dparser.mly"
                                  [] ) : trans_opt_type list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : trans_opt_type) in
    let _2 = (peek_val parser_env 0 : trans_opt_type list) in
    Obj.repr((
# 521 "src/dparser.mly"
                                  _1 :: _2 ) : trans_opt_type list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 524 "src/dparser.mly"
                         if _1 = "" then None else
  Some(verify_trigger_list _1) ) : 'trans_trigger))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 528 "src/dparser.mly"
                           _1 ) : 'optional_string))
; (fun parser_env ->
    Obj.repr((
# 529 "src/dparser.mly"
    "" ) : 'optional_string))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : 'trans_trigger) in
    let _3 = (peek_val parser_env 2 : 'optional_then) in
    let _4 = (peek_val parser_env 1 : trans_opt_type list) in
    let _5 = (peek_val parser_env 0 : Dlg.trans_next) in
    Obj.repr((
# 534 "src/dparser.mly"
    let result = { 
      Dlg.trans_trigger = _2 ; 
      Dlg.trans_str = None ;
      Dlg.action = None ;
      Dlg.journal_str = None;
      Dlg.unknown_flags = 0 ;
      Dlg.next = _5 ; 
    } in
    List.iter (fun elt -> match elt with
    | Trans_Reply(r) -> result.Dlg.trans_str <- r ;
    | Trans_Do(r) -> result.Dlg.action <- r ;
    | Trans_Journal(r) -> result.Dlg.journal_str <- r;
    | Trans_Flags(r) -> result.Dlg.unknown_flags <- r;
    ) _4 ;
    result 
  ) : Dlg.transition))
; (fun parser_env ->
    let _2 = (peek_val parser_env 4 : 'optional_string) in
    let _4 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _5 = (peek_val parser_env 1 : trans_opt_type list) in
    let _6 = (peek_val parser_env 0 : Dlg.trans_next) in
    Obj.repr((
# 551 "src/dparser.mly"
    let result = { 
      Dlg.trans_trigger = 
        if _2 = "" then None
        else Some(verify_trigger_list _2) ; 
      Dlg.trans_str = Some(_4) ;
      Dlg.action = None ;
      Dlg.journal_str = None;
      Dlg.unknown_flags = 0 ;
      Dlg.next = _6 ; 
    } in
    List.iter (fun elt -> match elt with
    | Trans_Reply(r) -> result.Dlg.trans_str <- r ;
    | Trans_Do(r) -> result.Dlg.action <- r ;
    | Trans_Journal(r) -> result.Dlg.journal_str <- r;
    | Trans_Flags(r) -> result.Dlg.unknown_flags <- r;
    ) _5 ;
    result 
  ) : Dlg.transition))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 570 "src/dparser.mly"
    { 
      Dlg.trans_trigger = None ; 
      Dlg.trans_str = None ;
      Dlg.action = None ;
      Dlg.journal_str = None;
      Dlg.unknown_flags = 0 ;
      Dlg.next = Dlg.Copy(_2,_3) ;
    } 
  ) : Dlg.transition))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 581 "src/dparser.mly"
                         Dlg.Symbolic(get_current_unit () ,_2) ) : Dlg.trans_next))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 582 "src/dparser.mly"
                    Dlg.Symbolic(get_current_unit () ,_2) ) : Dlg.trans_next))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 583 "src/dparser.mly"
                         Dlg.Symbolic(String.uppercase _2,_3) ) : Dlg.trans_next))
; (fun parser_env ->
    Obj.repr((
# 584 "src/dparser.mly"
                         Dlg.Exit ) : Dlg.trans_next))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : string) in
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 588 "src/dparser.mly"
    let result = Dlg.Local_String({ lse_male = _1 ; lse_male_sound = _2; 
                    lse_female = _1; lse_female_sound = _2;}) in 
    (match !Dlg.local_string_ht with
      Some(l) -> if not (List.mem result l ) then 
          Dlg.local_string_ht := Some(result :: l )
    | _ -> () ) ;
    result
  ) : Dlg.tlk_string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 3 : string) in
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 597 "src/dparser.mly"
    let result = Dlg.Local_String({ lse_male = _1 ; lse_male_sound = _2; 
          lse_female = _3; lse_female_sound = _4; }) in 
    (match !Dlg.local_string_ht with
      Some(l) -> if not (List.mem result l) then 
        Dlg.local_string_ht := Some(result :: l)
    | _ -> () ) ;
    result
  ) : Dlg.tlk_string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 605 "src/dparser.mly"
               Dlg.TLK_Index(_1) ) : Dlg.tlk_string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : int) in
    Obj.repr((
# 606 "src/dparser.mly"
               Dc.resolve_string_while_loading (Dlg.Trans_String(_1)) ) : Dlg.tlk_string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : int) in
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 608 "src/dparser.mly"
    let _ = Dc.set_string_while_loading _1 _2 in Dlg.TLK_Index(_1) ) : Dlg.tlk_string))
; (fun parser_env ->
    Obj.repr((
# 611 "src/dparser.mly"
                          "" ) : string))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 612 "src/dparser.mly"
                          _1 ) : string))
; (fun parser_env ->
    Obj.repr((
# 615 "src/dparser.mly"
                          [] ) : (int * Dlg.tlk_string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 3 : int) in
    let _3 = (peek_val parser_env 1 : Dlg.tlk_string) in
    let _4 = (peek_val parser_env 0 : (int * Dlg.tlk_string) list) in
    Obj.repr((
# 616 "src/dparser.mly"
                                  (_1,_3) :: _4 ) : (int * Dlg.tlk_string) list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 6 : string) in
    let _4 = (peek_val parser_env 4 : string) in
    let _5 = (peek_val parser_env 3 : Tp.tp_flag list) in
    let _6 = (peek_val parser_env 2 : 'allow_missing) in
    let _7 = (peek_val parser_env 1 : Tp.tp_lang list) in
    let _8 = (peek_val parser_env 0 : Tp.tp_mod list) in
    Obj.repr((
# 621 "src/dparser.mly"
    { Tp.backup = _2;
      Tp.author = _4; 
      Tp.flags = _5 ;
      Tp.allow_missing = _6; 
      Tp.languages = _7 ;
      Tp.modules = _8 ;
  } ) : Tp.tp_file))
; (fun parser_env ->
    Obj.repr((
# 630 "src/dparser.mly"
                                  [] ) : Tp.tp_flag list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Tp.tp_flag list) in
    Obj.repr((
# 631 "src/dparser.mly"
                                         (Tp.Auto_Tra(_2)) :: _3 ) : Tp.tp_flag list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Tp.tp_flag list) in
    Obj.repr((
# 632 "src/dparser.mly"
                                         (Tp.Menu_Style(_2)) :: _3 ) : Tp.tp_flag list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Tp.tp_flag list) in
    Obj.repr((
# 633 "src/dparser.mly"
                                         (Tp.Ask_Every_Component) :: _2 ) : Tp.tp_flag list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : Tp.tp_action list) in
    let _4 = (peek_val parser_env 0 : Tp.tp_flag list) in
    Obj.repr((
# 634 "src/dparser.mly"
                                             (Tp.Always(_2)) :: _4 ) : Tp.tp_flag list))
; (fun parser_env ->
    Obj.repr((
# 637 "src/dparser.mly"
                                        [] ) : 'allow_missing))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 638 "src/dparser.mly"
                                        _2 ) : 'allow_missing))
; (fun parser_env ->
    Obj.repr((
# 641 "src/dparser.mly"
                                 [] ) : string list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : string) in
    let _2 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 642 "src/dparser.mly"
                              (String.uppercase _1) :: _2 ) : string list))
; (fun parser_env ->
    Obj.repr((
# 645 "src/dparser.mly"
                   [] ) : Tp.tp_lang list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : string list) in
    let _5 = (peek_val parser_env 0 : Tp.tp_lang list) in
    Obj.repr((
# 647 "src/dparser.mly"
    { Tp.lang_name = _2 ;
      Tp.lang_dir_name = _3 ;
      Tp.lang_tra_files = _4 ;
  } :: _5 ) : Tp.tp_lang list))
; (fun parser_env ->
    Obj.repr((
# 653 "src/dparser.mly"
                        [] ) : Tp.tp_mod list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _3 = (peek_val parser_env 1 : Tp.tp_action list) in
    let _4 = (peek_val parser_env 0 : Tp.tp_mod list) in
    Obj.repr((
# 655 "src/dparser.mly"
    { Tp.mod_name = _2 ; Tp.mod_parts = _3 ; } :: _4 ) : Tp.tp_mod list))
; (fun parser_env ->
    Obj.repr((
# 658 "src/dparser.mly"
                              [] ) : Tp.tp_action list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : Tp.tp_action) in
    let _2 = (peek_val parser_env 0 : Tp.tp_action list) in
    Obj.repr((
# 659 "src/dparser.mly"
                              _1 :: _2 ) : Tp.tp_action list))
; (fun parser_env ->
    Obj.repr((
# 662 "src/dparser.mly"
                                  [] ) : 'optional_using))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 663 "src/dparser.mly"
                                  _2 ) : 'optional_using))
; (fun parser_env ->
    Obj.repr((
# 666 "src/dparser.mly"
                      true ) : 'optional_plus))
; (fun parser_env ->
    Obj.repr((
# 667 "src/dparser.mly"
                      false ) : 'optional_plus))
; (fun parser_env ->
    Obj.repr((
# 670 "src/dparser.mly"
                      false ) : 'optional_glob))
; (fun parser_env ->
    Obj.repr((
# 671 "src/dparser.mly"
                      true ) : 'optional_glob))
; (fun parser_env ->
    let _2 = (peek_val parser_env 4 : 'optional_plus) in
    let _3 = (peek_val parser_env 3 : 'optional_glob) in
    let _4 = (peek_val parser_env 2 : (string * string) list) in
    let _5 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _6 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 676 "src/dparser.mly"
    Tp.TP_Copy(
      { Tp.copy_get_existing = false;
        Tp.copy_use_regexp = false; 
        Tp.copy_use_glob = _3; 
        Tp.copy_file_list = _4 ;
        Tp.copy_patch_list = _5 ;
        Tp.copy_constraint_list = _6;
        Tp.copy_backup = _2 ; } ) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : 'optional_plus) in
    let _3 = (peek_val parser_env 2 : (string * string) list) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 685 "src/dparser.mly"
    Tp.TP_Copy(
      { Tp.copy_get_existing = true;
        Tp.copy_use_regexp = false; 
        Tp.copy_use_glob = false;
        Tp.copy_file_list = _3 ;
        Tp.copy_patch_list = _4 ;
        Tp.copy_constraint_list = _5;
        Tp.copy_backup = _2 ; } ) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : 'optional_plus) in
    let _3 = (peek_val parser_env 2 : (string * string) list) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 694 "src/dparser.mly"
    Tp.TP_Copy(
      { Tp.copy_get_existing = true;
        Tp.copy_use_regexp = true; 
        Tp.copy_use_glob = false; 
        Tp.copy_file_list = _3 ;
        Tp.copy_patch_list = _4 ;
        Tp.copy_constraint_list = _5;
        Tp.copy_backup = _2 ; } ) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string list) in
    let _3 = (peek_val parser_env 0 : Tp.tp_patch list) in
    Obj.repr((
# 702 "src/dparser.mly"
                                          Tp.TP_CopyRandom(_2,_3) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 703 "src/dparser.mly"
                       Tp.TP_RandomSeed(my_int_of_string _2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string list) in
    let _3 = (peek_val parser_env 0 : 'optional_using) in
    Obj.repr((
# 704 "src/dparser.mly"
                                       Tp.TP_Compile(_2,_3) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 705 "src/dparser.mly"
                      Tp.TP_Mkdir(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 706 "src/dparser.mly"
                            Tp.TP_Require_File(_2,_3) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 707 "src/dparser.mly"
                           Tp.TP_Forbid_File(_2,_3) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 708 "src/dparser.mly"
                                      Tp.TP_Append(_2,_3,_4) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _3 = (peek_val parser_env 1 : string) in
    let _4 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 710 "src/dparser.mly"
      let lst = Str.split (Str.regexp "[ \t]+") _3 in
      let lst = List.map (fun elt -> if elt = "$" then "" else elt) lst in
      Tp.TP_Append_Col(_2,lst,_4) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : 'optional_using) in
    Obj.repr((
# 714 "src/dparser.mly"
      Tp.TP_Extend_Top(false,_2,_3,_4,_5) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : 'optional_using) in
    Obj.repr((
# 716 "src/dparser.mly"
      Tp.TP_Extend_Bottom(false,_2,_3,_4,_5) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : 'optional_using) in
    Obj.repr((
# 718 "src/dparser.mly"
      Tp.TP_Extend_Top(true,_2,_3,_4,_5) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : string) in
    let _3 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    let _5 = (peek_val parser_env 0 : 'optional_using) in
    Obj.repr((
# 720 "src/dparser.mly"
      Tp.TP_Extend_Bottom(true,_2,_3,_4,_5) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 721 "src/dparser.mly"
                   Tp.TP_At_Exit(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 722 "src/dparser.mly"
                               Tp.TP_At_Interactive_Exit(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 723 "src/dparser.mly"
                                    Tp.TP_At_Interactive_Uninstall(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 724 "src/dparser.mly"
                        Tp.TP_At_Uninstall(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 725 "src/dparser.mly"
                            Tp.TP_Add_Music(
    { Tp.music_name = _2; 
      Tp.music_file = _3; } ) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 728 "src/dparser.mly"
                          Tp.TP_String_Set(_2,_3) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 729 "src/dparser.mly"
                            Tp.TP_Uninstall_Now(_2,(my_int_of_string _3)) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 19 : string) in
    let _3 = (peek_val parser_env 18 : string) in
    let _4 = (peek_val parser_env 17 : string) in
    let _5 = (peek_val parser_env 16 : string) in
    let _6 = (peek_val parser_env 15 : string) in
    let _7 = (peek_val parser_env 14 : string) in
    let _8 = (peek_val parser_env 13 : string) in
    let _9 = (peek_val parser_env 12 : string) in
    let _10 = (peek_val parser_env 11 : string) in
    let _11 = (peek_val parser_env 10 : string) in
    let _12 = (peek_val parser_env 9 : string) in
    let _13 = (peek_val parser_env 8 : string) in
    let _14 = (peek_val parser_env 7 : string) in
    let _15 = (peek_val parser_env 6 : string) in
    let _17 = (peek_val parser_env 4 : Dlg.tlk_string) in
    let _19 = (peek_val parser_env 2 : Dlg.tlk_string) in
    let _21 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 748 "src/dparser.mly"
    Tp.TP_Add_Kit(
   { Tp.kit_name = _2 ;
    Tp.clasweap = _3 ; 
    Tp.weapprof = _4 ;
    Tp.abclasrq = _5 ;
    Tp.abclsmod = _6 ;
    Tp.abdcdsrq = _7 ;
    Tp.abdcscrq = _8 ;
    Tp.alignmnt = _9 ;
    Tp.dualclas = _10 ;
    Tp.ability_file = _11 ;
    Tp.include_in = _12 ;
    Tp.unused_class = _13 ;
    Tp.tob_abbr = _14; 
    Tp.tob_start = 
      (let lst = Str.split (Str.regexp "[ \t]+") _15 in
      List.map (fun elt -> if elt = "$" then "" else elt) lst) ;
    Tp.lower = _17 ;
    Tp.mixed = _19 ;
    Tp.help = _21 ;
  })) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 769 "src/dparser.mly"
                   Tp.TP_Fail(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 770 "src/dparser.mly"
                    Tp.TP_Print(_2) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 4 : 'tp_predicate) in
    let _5 = (peek_val parser_env 1 : Tp.tp_action list) in
    Obj.repr((
# 771 "src/dparser.mly"
                                                         Tp.TP_If(_2,_5,[]) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 8 : 'tp_predicate) in
    let _5 = (peek_val parser_env 5 : Tp.tp_action list) in
    let _9 = (peek_val parser_env 1 : Tp.tp_action list) in
    Obj.repr((
# 773 "src/dparser.mly"
                               Tp.TP_If(_2,_5,_9) ) : Tp.tp_action))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'tp_predicate) in
    Obj.repr((
# 776 "src/dparser.mly"
                                            _2 ) : 'tp_predicate))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'tp_predicate) in
    let _3 = (peek_val parser_env 0 : 'tp_predicate) in
    Obj.repr((
# 777 "src/dparser.mly"
                                  Tp.Pred_And(_1,_3) ) : 'tp_predicate))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'tp_predicate) in
    let _3 = (peek_val parser_env 0 : 'tp_predicate) in
    Obj.repr((
# 778 "src/dparser.mly"
                                 Tp.Pred_Or(_1,_3) ) : 'tp_predicate))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : 'tp_predicate) in
    Obj.repr((
# 779 "src/dparser.mly"
                     Tp.Pred_Not(_2) ) : 'tp_predicate))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 780 "src/dparser.mly"
                       Tp.Pred_File_Exists(_2) ) : 'tp_predicate))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 781 "src/dparser.mly"
                               Tp.Pred_File_Exists_In_Game(_2) ) : 'tp_predicate))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 782 "src/dparser.mly"
                            Tp.Pred_File_Size(_2,my_int_of_string _3) ) : 'tp_predicate))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 783 "src/dparser.mly"
                                Tp.Pred_File_Contains(_2,_3) ) : 'tp_predicate))
; (fun parser_env ->
    Obj.repr((
# 787 "src/dparser.mly"
                           [] ) : string list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : string) in
    let _2 = (peek_val parser_env 0 : string list) in
    Obj.repr((
# 788 "src/dparser.mly"
                        _1 :: _2 ) : string list))
; (fun parser_env ->
    Obj.repr((
# 791 "src/dparser.mly"
                                 [] ) : Tp.tp_constraint list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 792 "src/dparser.mly"
                                 Tp.TP_Contains(_2) :: _3 ) : Tp.tp_constraint list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 793 "src/dparser.mly"
                                 Tp.TP_NotContains(_2) :: _3 ) : Tp.tp_constraint list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 794 "src/dparser.mly"
                                   Tp.TP_IfSizeIs(my_int_of_string _2) :: _3 ) : Tp.tp_constraint list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : Tp.tp_constraint list) in
    Obj.repr((
# 795 "src/dparser.mly"
                                   Tp.TP_Eval(_2) :: _3 ) : Tp.tp_constraint list))
; (fun parser_env ->
    Obj.repr((
# 798 "src/dparser.mly"
                               [] ) : (string * string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : string) in
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : (string * string) list) in
    Obj.repr((
# 799 "src/dparser.mly"
                               (_1,_2) :: _3 ) : (string * string) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 802 "src/dparser.mly"
                                  Tp.PE_String(_1) ) : 'patch_exp))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    Obj.repr((
# 803 "src/dparser.mly"
                                  _2 ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 804 "src/dparser.mly"
                                  Tp.PE_Add(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 805 "src/dparser.mly"
                                   Tp.PE_Sub(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 806 "src/dparser.mly"
                                   Tp.PE_Mul(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 807 "src/dparser.mly"
                                    Tp.PE_Div(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 808 "src/dparser.mly"
                                Tp.PE_Equal(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _2 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 809 "src/dparser.mly"
                                Tp.PE_Not(_2) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 810 "src/dparser.mly"
                             Tp.PE_And(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 811 "src/dparser.mly"
                            Tp.PE_Or(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 812 "src/dparser.mly"
                            Tp.PE_GT(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 813 "src/dparser.mly"
                             Tp.PE_GTE(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 814 "src/dparser.mly"
                            Tp.PE_LT(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 2 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 815 "src/dparser.mly"
                             Tp.PE_LTE(_1,_3) ) : 'patch_exp))
; (fun parser_env ->
    let _1 = (peek_val parser_env 4 : 'patch_exp) in
    let _3 = (peek_val parser_env 2 : 'patch_exp) in
    let _5 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 816 "src/dparser.mly"
                                                 Tp.PE_If(_1,_3,_5); ) : 'patch_exp))
; (fun parser_env ->
    Obj.repr((
# 819 "src/dparser.mly"
                    [] ) : Tp.tp_patch list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 1 : 'tp_patch) in
    let _2 = (peek_val parser_env 0 : Tp.tp_patch list) in
    Obj.repr((
# 820 "src/dparser.mly"
                           _1 :: _2 ) : Tp.tp_patch list))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 824 "src/dparser.mly"
                      Tp.TP_PatchStrRef(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : Dlg.tlk_string) in
    Obj.repr((
# 825 "src/dparser.mly"
                       Tp.TP_PatchString(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 826 "src/dparser.mly"
                                    Tp.TP_PatchStringTextually(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : string) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 827 "src/dparser.mly"
                                    Tp.TP_PatchReplaceBCSBlock(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 828 "src/dparser.mly"
                               Tp.TP_PatchReadByte(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 829 "src/dparser.mly"
                                Tp.TP_PatchReadShort(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 830 "src/dparser.mly"
                               Tp.TP_PatchReadLong(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 831 "src/dparser.mly"
                                   Tp.TP_PatchByte(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 832 "src/dparser.mly"
                                    Tp.TP_PatchShort(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 833 "src/dparser.mly"
                                   Tp.TP_PatchLong(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 834 "src/dparser.mly"
                                 Tp.TP_PatchASCII(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 835 "src/dparser.mly"
                                     Tp.TP_PatchInsertBytes(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 1 : 'patch_exp) in
    let _3 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 836 "src/dparser.mly"
                                     Tp.TP_PatchDeleteBytes(_2,_3) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 2 : string) in
    let _4 = (peek_val parser_env 0 : 'patch_exp) in
    Obj.repr((
# 837 "src/dparser.mly"
                                Tp.TP_PatchSet(_2,_4) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : 'patch_exp) in
    let _4 = (peek_val parser_env 1 : Tp.tp_patch list) in
    Obj.repr((
# 838 "src/dparser.mly"
                                            Tp.TP_PatchWhile(_2,_4) ) : 'tp_patch))
; (fun parser_env ->
    let _2 = (peek_val parser_env 3 : 'patch_exp) in
    let _3 = (peek_val parser_env 2 : 'patch_exp) in
    let _4 = (peek_val parser_env 1 : 'patch_exp) in
    let _5 = (peek_val parser_env 0 : string) in
    Obj.repr((
# 839 "src/dparser.mly"
                                                       Tp.TP_Patch2DA(_2,_3,_4,_5) ) : 'tp_patch))
; (fun parser_env ->
    let _1 = (peek_val parser_env 0 : (string * int * int) list) in
    Obj.repr((
# 842 "src/dparser.mly"
                                _1 ) : (string * int * int) list))
; (fun parser_env ->
    Obj.repr((
# 845 "src/dparser.mly"
                                [] ) : (string * int * int) list))
; (fun parser_env ->
    let _1 = (peek_val parser_env 3 : string) in
    let _2 = (peek_val parser_env 2 : int) in
    let _3 = (peek_val parser_env 1 : int) in
    let _4 = (peek_val parser_env 0 : (string * int * int) list) in
    Obj.repr((
# 846 "src/dparser.mly"
                                                    (_1,_2,_3) :: _4 ) : (string * int * int) list))
(* Entry d_file *)
; (fun parser_env -> raise (YYexit (peek_val parser_env 0)))
(* Entry tra_file *)
; (fun parser_env -> raise (YYexit (peek_val parser_env 0)))
(* Entry tp_file *)
; (fun parser_env -> raise (YYexit (peek_val parser_env 0)))
(* Entry log_file *)
; (fun parser_env -> raise (YYexit (peek_val parser_env 0)))
|]
let yytables =
  { actions=yyact;
    transl_const=yytransl_const;
    transl_block=yytransl_block;
    lhs=yylhs;
    len=yylen;
    defred=yydefred;
    dgoto=yydgoto;
    sindex=yysindex;
    rindex=yyrindex;
    gindex=yygindex;
    tablesize=yytablesize;
    table=yytable;
    check=yycheck;
    error_function=parse_error;
    names_const=yynames_const;
    names_block=yynames_block }
let d_file (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (yyparse yytables 1 lexfun lexbuf : Dc.action list)
let tra_file (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (yyparse yytables 2 lexfun lexbuf : (int * Dlg.tlk_string) list)
let tp_file (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (yyparse yytables 3 lexfun lexbuf : Tp.tp_file)
let log_file (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (yyparse yytables 4 lexfun lexbuf : (string * int * int) list)
(* Line 850, file src/dparser.mly *)



(* Line 2737, file src/dparser.ml *)
